/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

/* Documentation Channel Modul DC: XPLAB's internal data documentation 
   feature. The DC is a database application and stores and produces 
   different data structures needed for output. The motivation is to separate
   code for documentation and code for algorithms. The algorithms don't need
   any part of the DC.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_dchannel.h"
#include "xplab_image.h"
#include "xplab_algorithm.h"
#include "xplab_report.h"
#include "xplab_preferences_fe.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_batch.h"
#include "xplab_tprint.h"

/* local definitions *******************************************************/
typedef struct {
  int     styp;                     /* source flag                          */
  int     req;                      /* required flag                        */
  char    *tname;                   /* filename of (temporary) stored image */
  pic_def *tpic;                    /* buffer for temporary stored image    */
} dc_pic_src;                       /* definition image source              */

typedef struct {
  area_def   sar;             /* general source attribute image size        */
  int        store_size;      /* no. of images to store internally          */
  int        store_sint;      /* store sources internally (1) or extern (0) */
  int        cnt;             /* array size                                 */
  dc_pic_src *dps;            /* source array                               */
} dc_pic_lst;                 /* definition image source list               */

typedef struct {
  int2_lst   opt_1d;               /* DC option list for 1-D data           */
  int2_lst   opt_2d;               /* DC option list for 2-D data           */
  dc_pic_lst dpl;                  /* image source management               */
} dc_pic;                          /* definition data documentation channel */

/* local prototypes ********************************************************/
static void init_dc_pic_src(dc_pic_src *dps);
static void init_dc_pic_lst(dc_pic_lst *dpl);
static void init_dc_pic(dc_pic *dp);
static void copy_dc_pic_src(dc_pic_src *src, dc_pic_src *dst);
static void move_dc_pic_src(dc_pic_src *src, dc_pic_src *dst);
static void delete_dc_pic_src(dc_pic_lst *dpl, int idx);
static void free_dc_pic(dc_pic *dp);
static void free_dc_pic_lst(dc_pic_lst *dpl);
static void free_dc_pic_src(dc_pic_src *dps);

static void dc_store_source(char *dc_name, pic_def src_pic, int src_type, 
		     dc_pic *dcp);
static void dc_store_all(char *dc_name, rb_lst src_rb, rb_lst src_mrb, 
		  cand_lst src_cand, cls_def *src_cls, int_lst src_hline, 
		  dc_pic *dcp);
static void dc_store_pic(char *dc_name, pic_def src_pic, int src_type, dc_pic *dcp);
static void dc_store_nb(char *dc_name, cand_lst src_cand, cls_def *src_cls, dc_pic *dcp);
static void dc_store_rb(char *dc_name, rb_lst src_rb, dc_pic *dcp);
static void dc_store_mrb(char *dc_name, rb_lst src_mrb, dc_pic *dcp);
static void dc_store_pattern(char *dc_name, cand_lst src_cand, dc_pic *dcp);
static void dc_store_hline(char *dc_name, int_lst src_hline, dc_pic *dcp);
static void dc_store_pic_dbl(char *dc_name, pic_dbl source_dpic, int source_type, 
		      dc_pic *dcp);
static void dc_update(pic_def psum, dc_pic_lst *dpl, int flag_list, char *tname);
static void dc_overlay(pic_def src, pic_def *dst, col_def foreground_color);
static SRC_IS dc_overlay_all(rb_lst src_rb, rb_lst src_mrb, cand_lst src_cand, 
		      cls_def *src_cls, int_lst src_hline, char *fname,
		      dc_pic_lst *dpl, int *flag_list, int preview);
static SRC_IS dc_overlay_pic(pic_def src_pic, int src_type, char *fname, 
		      dc_pic_lst *dpl, int *flag_list, int preview);
static SRC_IS dc_overlay_rb(rb_lst src_rb, char *fname, dc_pic_lst *dpl, 
		     int *flag_list, int preview);
static SRC_IS dc_overlay_mrb(rb_lst src_mrb, char *fname, dc_pic_lst *dpl, 
		      int *flag_list, int preview);
static SRC_IS dc_overlay_pattern(cand_lst src_cand, char *fname, dc_pic_lst *dpl,
		          int *flag_list, int preview);
static SRC_IS dc_overlay_nb(cand_lst src_cand, cls_def *src_cls, char *fname, 
		     dc_pic_lst *dpl, int *flag_list, int preview);
static SRC_IS dc_overlay_hline(int_lst src_hline, char *fname, dc_pic_lst *dpl, 
		        int *flag_list, int preview);
static SRC_IS dc_composition(pic_def src_pic, int src_type, rb_lst src_rb, 
		      rb_lst src_mrb, cand_lst src_cand, cls_def *src_cls, 
		      int_lst src_hline, char *fname, dc_pic_lst *dpl, 
		      int *flag_list, int preview);
static SRC_IS dc_overlay_method(cand_lst src_cand, dc_pic_lst dpl, int *flag_list, 
		         pic_def *psum, char **tname, int preview);
static SRC_IS dc_overlay_method_rb(cand_lst src_cand, PATTERN_ATT pattern_flag, 
			    char *ext, dc_pic_lst dpl, int flag_list, 
			    PIXEL_KEY grey_backgnd, pic_def *psum, 
			    char **tname, int preview);
static SRC_IS dc_compose_overlay(pic_def src_pic, int src_type, rb_lst src_rb, 
		          rb_lst src_mrb, cand_lst src_cand, 
			  cls_def *src_cls, int_lst src_hline, char *fname,
			  dc_pic_lst dpl, int *flag_list, pic_def *psum,
			  char **tname, int preview);
static SRC_IS dc_compose_dot(pic_def src_pic, int src_flag, dc_pic_lst dpl, 
		      pic_def *psum, char **tname, int preview);
static SRC_IS dc_compose_doc(pic_def src_pic, int src_type, dc_pic_lst dpl, 
		      pic_def *psum, char **tname, int preview);
static SRC_IS dc_compose_gv(pic_def src_pic, int src_type, dc_pic_lst dpl, 
		     pic_def *psum, char **tname, int preview);
static SRC_IS dc_compose_clean(pic_def src_pic, int src_type, dc_pic_lst dpl, 
	  	        pic_def *psum, char **tname, int preview);
static SRC_IS dc_compose_org(pic_def src_pic, int src_type, dc_pic_lst dpl, 
		      pic_def *psum, char **tname, int preview);
static SRC_IS dc_compose_frame(rb_lst src_rb, pic_def *psum, char *ext,
		        col_def *frame_color, char **tname, int preview);
static SRC_IS dc_compose_rb(cand_lst hlist, PATTERN_ATT pattern_flag, 
		     char *ext_method, char *ext_src, dc_pic_lst dpl, 
		     int src_flag, PIXEL_KEY grey_backgnd, pic_def *psum, 
		     char **tname, int preview);
static SRC_IS dc_compose_nb(cand_lst hlist, cls_def *cls, dc_pic_lst dpl, 
		     pic_def *psum, char **tname, int preview);
static SRC_IS dc_compose_hline(int_lst hline, dc_pic_lst dpl, pic_def *psum, 
			char *ext, col_def *line_color, char **tname, 
			int preview);
static SRC_IS dc_compose_sline(rb_lst src_rb, pic_def *psum, char *ext, 
		        col_def *line_color, char **tname, int preview);
static int dc_dependence(int2_lst opt, int src_type);
static pic_def dc_load_source(dc_pic_lst dpl, int sidx);
static int dc_exist_name(dc_pic_lst dpl, char *name);
static int dc_exist_required(dc_pic_lst dpl, int flag);
static int dc_exist_source(dc_pic_lst dpl, int flag);
static char *dc_base_name(char *filename, char *type);
static int dc_reset(dc_pic *dcp);
static void dc_delete_source(dc_pic_lst *dpl, int sidx);
static int dc_delete_flags(int2_lst dol);
static pic_def dc_copy_grid(rb_lst rbl, col_def *color);
static pic_def dc_copy_frames(rb_lst rbl, col_def *color);
static pic_def dc_copy_lines(area_def ar_in, col_def *color, int_lst line);
static pic_def dc_cinderella_nb(area_def doc_ar, cand_lst hlist, cls_def *cls);
static pic_def dc_cinderella_pattern(pic_def pic, cand_lst hlist, PATTERN_ATT att,
			      PIXEL_KEY grey_backgnd);
static SRC_IS dc_1d_compose_ccf(fft_def src_flst, int src_type, double_lst *fsum, 
		         char **tname, char **comment, int preview);
static SRC_IS dc_1d_compose_sigin(double_lst src_fctn, int src_type, 
			   double_lst *fsum, char **tname, char **comment, 
			   int preview);
static SRC_IS dc_1d_compose_pos(double_lst src_fctn, int src_type, double_lst *fsum, 
		         char **tname, char **comment, int preview);
static SRC_IS dc_1d_compose_hphase(double_lst src_fctn, int src_type, 
			    double_lst *fsum, char **tname, char **comment, 
			    int preview);
static SRC_IS dc_1d_compose_tfin(double_lst src_fctn, int src_type, 
			  double_lst *fsum, char **tname, char **comment, 
			  int preview);
static SRC_IS dc_1d_compose_tflag(double_lst src_fctn, int src_type, 
			   double_lst *fsum, char **tname, char **comment, 
			   int preview);
static SRC_IS dc_1d_compose(double_lst src_fctn, fft_def src_fft, int source_type, 
		     char *fname, int *flag_list, double_lst *fsum, 
		     char **tname, char **comment, int preview);
static SRC_IS dc_1d_composition(double_lst src_fctn, fft_def src_fft, int src_type,
		         char *fname, char *subject, int *flag_list, 
		         int preview);
static void dc_1d_store_double(char *dc_name, double_lst src_fctn, int source_type,
			char *subject, dc_pic *dcp);
static SRC_IS dc_1d_ccf(fft_def src_fft, int src_type, char *fname, char *subject, 
	         int *flag_list, int preview);
static SRC_IS dc_1d_double(double_lst src_fctn, int src_type, char *fname, 
		    char *subject, int *flag_list, int preview);
static void dc_1d_store_fft(char *dc_name, fft_def src_fft, int source_type, 
		     char *subject, dc_pic *dcp);
static SRC_IS dc_compose_ccf(pic_dbl dpic, char *ext_src, pic_def *psum, 
		      char **tname, int preview);
static SRC_IS dc_direct(pic_dbl src_dpic, int src_type, char *fname, 
		 int *flag_list, int preview);
static SRC_IS dc_direct_pic_dbl(pic_dbl src_dpic, int src_type, char *fname, 
			 int *flag_list, int preview);
static void dc_store_type(dc_pic *dcp);
static int check_dc_name(char *dcname, char *desc);
static void xdc_assemble_sample(void);
static dc_pic *xdc_get_channel(void);

/* Functions **************************************************************/

static void 
init_dc_pic_lst(dc_pic_lst *dpl)
{
  assert(dpl);

  dpl->sar.width  = 0;
  dpl->sar.height = 0;
  dpl->store_size = 0;
  dpl->store_sint = 0;
  dpl->cnt = 0;
  dpl->dps = 0;                    /* zero pointer */
}


static void 
init_dc_pic_src(dc_pic_src *dps)
{
  assert(dps);

  dps->styp = 0;
  dps->req = 0;
  dps->tname = 0;                    /* zero pointer */
  dps->tpic = 0;                     /* zero pointer */
}


static void 
init_dc_pic(dc_pic *dcp)
{
  assert(dcp);

  init_int2_lst(&dcp->opt_1d);
  init_int2_lst(&dcp->opt_2d);
  init_dc_pic_lst(&dcp->dpl);
}


static void 
copy_dc_pic_src(dc_pic_src *src, dc_pic_src *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  free_dc_pic_src(dst);

  dst->styp = src->styp;
  dst->req = src->req;

  add_string(src->tname, &dst->tname);

  CALLOC(pic_def, 1, dst->tpic);

  if (src->tpic) {                               /* pointer may be cleared */
    copy_pic_def(src->tpic, dst->tpic);
  }
}


static void 
move_dc_pic_src(dc_pic_src *src, dc_pic_src *dst) 
{   
  assert(src);
  assert(dst);
  assert(src != dst);

  copy_dc_pic_src(src, dst);

  free_dc_pic_src(src);
}


static void
delete_dc_pic_src(dc_pic_lst *dpl, int idx)
{
  int last;

  assert(dpl);
  assert(idx>=0);
  assert(idx<dpl->cnt);

  if (!dpl->cnt) {                                    /* consistency check */
    assert(!dpl->dps);
    return;
  }
  else {
    assert(dpl->cnt>0);
    assert(dpl->dps);
  }

  last = dpl->cnt-1;

  if (idx == last) {
    free_dc_pic_src(&dpl->dps[idx]);

    if (!last) {
      FREEMEM(dpl->dps);
    }
  }
  else {
    move_dc_pic_src(&dpl->dps[last], &dpl->dps[idx]);
  }

  dpl->cnt = last;
}


static void
free_dc_pic_src(dc_pic_src *dps)
{
  assert(dps);

  FREEMEM(dps->tname);

  if (dps->tpic) {                               /* pointer may be cleared */
    free_pic_def(dps->tpic);
    FREEMEM(dps->tpic);
  }

  init_dc_pic_src(dps);
}


static void
free_dc_pic_lst(dc_pic_lst *dpl)
{
  int i;

  assert(dpl);

  if (!dpl->cnt) {                                    /* consistency check */
    assert(!dpl->dps);
  }
  else {
    assert(dpl->cnt>0);
    assert(dpl->dps);
  }

  for (i = 0; i < dpl->cnt ; i++) {
    free_dc_pic_src(&dpl->dps[i]);
  }
  FREEMEM(dpl->dps);

  init_dc_pic_lst(dpl);
}


static void
free_dc_pic(dc_pic *dp)
{
  assert(dp);

  free_int2_lst(&dp->opt_1d);
  free_int2_lst(&dp->opt_2d);
  free_dc_pic_lst(&dp->dpl);

  init_dc_pic(dp);
}



/* Deletes all options in an option list.

Revision: 30.12.2002 E.Sack

Input : - option list
Output: - option list updated
        - number of deleted options
*/
static int
dc_delete_flags(int2_lst dol)
{
  int i, del = 0;

  for (i = 0; i < dol.cnt; i++) {
    if (dol.n[i].nr) {
      dol.n[i].nr = 0;
      del++;
    }
  }

  return del;
}



/* Searches for a required composition.

Revision: 29.12.2002 E.Sack

Input : - image list
        - composition
Output: - index (-1: no composition found)
*/
static int
dc_exist_required(dc_pic_lst dpl, int flag)
{
  int i;

  for (i = 0; i < dpl.cnt; i++) {
    if (flag == dpl.dps[i].req) {
      return i;
    }
  }

  return -1;
}



/* Searches for a documentation channel source.

Revision: 27.12.2002 E.Sack

Input : - image list
        - image source flag
Output: - source index (-1: no source found)
*/
static int
dc_exist_source(dc_pic_lst dpl, int flag)
{
  int i;

  for (i = 0; i < dpl.cnt; i++) {
    if (flag == dpl.dps[i].styp) {
      return i;
    }
  }

  return -1;
}



/* Loads an image source from stock.

Revision: 28.12.2002 E.Sack

Input : - image list
        - source index
        - preview
Output: - image
*/
static pic_def
dc_load_source(dc_pic_lst dpl, int sidx)
{
  int     mgv;
  pic_def source;

  assert(sidx >= 0);
  assert(sidx<dpl.cnt);

  init_pic_def(&source);

  if (dpl.dps[sidx].tname) {
    assert(!dpl.dps[sidx].tpic);                  /* only one source store */
    source = read_pnm_file(dpl.dps[sidx].tname, &mgv, 
			   xctrl.cfg.verbose >= VERBOSE_LOT).pimg;
  }
  else {
    assert(dpl.dps[sidx].tpic);                   /* only one source store */
    assert(!dpl.dps[sidx].tname);                 /* only one source store */
    copy_pic_def(dpl.dps[sidx].tpic, &source);              /* copy source */
  }

  return source;
}



/* Deletes an image from a documentation channel.

Revision: 30.12.2002 E.Sack

Input : - pointer to image list
        - index
Output: - image list updated
*/
static void
dc_delete_source(dc_pic_lst *dpl, int sidx)
{
  int d;

  assert(sidx>=0);
  assert(sidx<dpl->cnt);
  assert(!dpl->dps[sidx].req);                             /* not required */

  if (dpl->dps[sidx].tname) {
    assert(!dpl->dps[sidx].tpic);                 /* only one source store */

    d = remove(dpl->dps[sidx].tname);             /* remove file from disk */

    if (d) {
      fprintf(stderr, "\nERROR: couldn't remove '%s' from disk\n\n",
	      dpl->dps[sidx].tname);
    }
  }
  else {
    assert(dpl->dps[sidx].tpic);                  /* only one source store */
    assert(!dpl->dps[sidx].tname);                /* only one source store */
  }

  delete_dc_pic_src(dpl, sidx);
}



/*  Checks backward dependence of the source.

Revision: 30.12.2002 E.Sack

Input : - overlay flags list
        - source type
Output: - flag (0: source type not required, 1: source type required)
*/
static int
dc_dependence(int2_lst opt, int src_type)
{
  int i, stype;

  for (i = 0; i < opt.cnt; i++) {       /* 1.condition: direct use of type */
    if (opt.n[i].nr == src_type) {                           /* check flag */
      return 1;                                              /* used alone */
    }
  }

  for (i = 0; i < opt.cnt; i++) {    /* 2.condition: indirect use by comp. */
    if (opt.n[i].nr & src_type) {                            /* check flag */
      return 1;                                   /* used in a composition */
    }
  }

  switch (src_type) {           /* 3.condition: indirect use by dependence */
  case DC_BIT_ORG:
    break;
  case DC_BIT_BMP:                                /* source used by others */
    stype = dc_dependence(opt, DC_BIT_DOT);
    return stype;
    break;
  case DC_BIT_DOT:
    break;
  case DC_BIT_NB_GOOD:
    break;
  case DC_BIT_RB_GOOD:
    break;
  case DC_BIT_RB_FAIL:
    break;
  case DC_BIT_CLEAN:                              /* source used by others */
    stype = dc_dependence(opt, DC_BIT_DOT);
    return stype;
    break;
  case DC_BIT_MRB_FRAME_BMP:
    break;
  case DC_BIT_MRB_FRAME_RGB:
    break;
  case DC_BIT_RB_FRAME_BMP:
    break;
  case DC_BIT_RB_FRAME_RGB:
    break;
  case DC_BIT_HLINE_BMP:
    break;
  case DC_BIT_HLINE_RGB:
    break;
  case DC_BIT_GV:
    break;
  default: 
    fprintf(stderr, "\nERROR: dependence missing for '0x%X'\n\n", src_type);
    assert(0);                                          /* service missing */
    break;
  }

  return 0;
}



/* Makes an overlay of two images.

Revision: 28.12.2002 E.Sack

Input : - source image
        - pointer to destination image
        - foreground color
Output: - destination image updated
Precondition: - if destination isn't allocated, it will be allocated.
              - foreground color needed for bitmap to color conversion
*/
static void
dc_overlay(pic_def src, pic_def *dst, col_def foreground_color)
{
  int src_is_bitmap, dst_is_bitmap, src_is_grey, dst_is_grey;
  int src_is_color, dst_is_color;
  PNM_FORMAT type_src, type_dst;
  pic_def    ptemp, ptemp2;

  assert(dst);

  init_pic_def(&ptemp);
  init_pic_def(&ptemp2);

  if (!dst->mem && !dst->col) {
    copy_pic_def(&src, dst);
    return;
  }

  assert(src.ar.width == dst->ar.width);
  assert(src.ar.height == dst->ar.height);
  assert((src.mem&&!src.col) || (!src.mem&&src.col));
  assert((dst->mem&&!dst->col) || (!dst->mem&&dst->col));

  type_src = automatic_format(src, src.ar, 0, 0, xctrl.cfg.rawbit);
  type_dst = automatic_format(*dst, dst->ar, 0, 0, xctrl.cfg.rawbit);

  assert(type_src!=PNM_FORMAT_UNDEF && type_dst!=PNM_FORMAT_UNDEF);
  assert(type_src!=PNM_FORMAT_MIXED && type_dst!=PNM_FORMAT_MIXED);

  src_is_bitmap = type_src == P4 || type_src == P1;
  dst_is_bitmap = type_dst == P4 || type_dst == P1;

  src_is_grey = type_src == P5 || type_src == P2;
  dst_is_grey = type_dst == P5 || type_dst == P2;

  src_is_color = type_src == P6 || type_src == P3;
  dst_is_color = type_dst == P6 || type_dst == P3;

  if (src_is_bitmap && dst_is_bitmap) {
    add_bmps(src, *dst, src.ar, 0, 0, 0, 0);
    return;
  }

  if (src_is_grey && dst_is_grey) {
    add_grey_norm(src, *dst, src.ar, 0, 0, 0, 0);
    return;
  }

  if (src_is_color && dst_is_color) {
    add_color_norm(src, *dst, src.ar, 0, 0, 0, 0);
    return;
  }

  if (src_is_bitmap && dst_is_grey) {
    add_bitmap_to_grey(src, *dst, src.ar, 0, 0, 0, 0);
    return;
  }

  if (src_is_grey && dst_is_color) {
    ptemp = grey_to_color(src);
    add_color_norm(ptemp, *dst, ptemp.ar, 0, 0, 0, 0);
    free_pic_def(&ptemp);
    return;
  }

  if (src_is_grey && dst_is_bitmap) {
    copy_pic_def(&src, &ptemp);
    add_bitmap_to_grey(*dst, ptemp, dst->ar, 0, 0, 0, 0);
    copy_pic_def(&ptemp, dst);
    free_pic_def(&ptemp);
    return;
  }

  if (src_is_color && dst_is_grey) {
    ptemp = grey_to_color(*dst);
    add_color_norm(src, ptemp, src.ar, 0, 0, 0, 0);
    copy_pic_def(&ptemp, dst);
    free_pic_def(&ptemp);
    return;
  }

  if (src_is_bitmap && dst_is_color) {
    add_bitmap_to_color(src, *dst, src.ar, 0, 0, 0, 0, foreground_color);
    return;
  }

  if (src_is_color && dst_is_bitmap) {
    copy_pic_def(&src, &ptemp);
    add_bitmap_to_color(*dst, ptemp, dst->ar, 0, 0, 0, 0, foreground_color);
    copy_pic_def(&ptemp, dst);
    free_pic_def(&ptemp);
    return;
  }

  assert(0);                                         /* conversion missing */
}



/* Builds image of discarded pixel. A preview of source completion is 
   possible which avoids unnecessary computation.

Revision: 29.12.2002 E.Sack

Input : - source image
        - source image type
        - image list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_dot(pic_def src_pic, int src_type, dc_pic_lst dpl, 
	       pic_def *psum, char **tname, int preview)
{
  int     idx_org = -1, idx_rb = -1;
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic_org, pic_rb;

  assert(tname);

  init_pic_def(&pic_org);
  init_pic_def(&pic_rb);

  if (src_type != DC_BIT_BMP) {                       /* source dependence */
    idx_org = dc_exist_source(dpl, DC_BIT_BMP);

    if (idx_org < 0) {                               /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else if (!preview) {
      pic_org = dc_load_source(dpl, idx_org);
    }
  }
  else if (!preview) {                              /* direct input source */
    copy_pic_def(&src_pic, &pic_org);
  }

  if (source_available == SRC_VALID_COMPLETE) {
    if (src_type != DC_BIT_CLEAN) {                   /* source dependence */
      idx_rb  = dc_exist_source(dpl, DC_BIT_CLEAN);

      if (idx_rb < 0) {                              /* check availability */
	source_available = SRC_VALID_INCOMPLETE;   /* source not available */
      }
      else if (!preview) {
	pic_rb = dc_load_source(dpl, idx_rb);
      }
    }
    else if (!preview) {                            /* direct input source */
      copy_pic_def(&src_pic, &pic_rb);
    }

    if (!preview && source_available == SRC_VALID_COMPLETE) {
      col_def foreground_color = FGND_COLOR;

      mark_diff(pic_rb, pic_org, pic_rb.ar, 0, 0, 0, 0, FOREGROUND);

      dc_overlay(pic_org, psum, foreground_color);

      add_string(DC_OPT_DOT_FILE, tname);
    }
  }

  free_pic_def(&pic_org);
  free_pic_def(&pic_rb);

  return source_available;
}



/* Builds raw document image source. A preview of source completion is 
   possible which avoids unnecessary computation.

Revision: 30.12.2002 E.Sack

Input : - source image
        - source image type
        - image list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_doc(pic_def src_pic, int src_type, dc_pic_lst dpl, pic_def *psum,
	       char **tname, int preview)
{
  int     idx_src = -1;
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic;

  assert(tname);

  init_pic_def(&pic);

  if (src_type != DC_BIT_ORG) {                       /* source dependence */
    idx_src = dc_exist_source(dpl, DC_BIT_ORG);             /* check stock */

    if (idx_src < 0) {                               /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else if (!preview) {
      pic = dc_load_source(dpl, idx_src);
    }
  }
  else if (!preview) {                              /* direct input source */
    copy_pic_def(&src_pic, &pic);
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    dc_overlay(pic, psum, foreground_color);

    add_string(DC_OPT_ORG_FILE, tname);
  }

  free_pic_def(&pic);

  return source_available;
}



/* Dumps samples. A preview of source completion is possible which avoids
   unnecessary computation.

Revision: 06.01.2003 E.Sack

Input : - source FFT
        - source type
        - pointer to output sample list
        - pointer to samples name
        - pointer to comment string
        - preview
Output: - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_compose_ccf(fft_def src_fft, int src_type, double_lst *fsum, 
		  char **tname, char **comment, int preview)
{
  SRC_IS source_available = SRC_INVALID;

  assert(tname);
  assert(comment);

  if (src_type == DC_BIT_POS_CCF) {                   /* source dependence */
    if (!src_fft.data || !src_fft.ar.width) {        /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else {
      assert(src_fft.data[0]);
      source_available = SRC_VALID_COMPLETE;           /* source available */

      if (!preview) {
	add_string(DC_OPT_POS_CCF_FILE, tname);
	cat_string(DC_PREFIX, DC_OPT_POS_CCF_DESC, DC_SUFFIX, comment);

	assert(!fsum->cnt && !fsum->dbl);

	fsum->cnt = src_fft.ar.width;
	CALLOC(double, fsum->cnt, fsum->dbl);

	copy_double(src_fft.data[0], fsum->dbl, src_fft.ar.width);
      }
    }
  }

  return source_available;
}



/* Dumps samples. A preview of source completion is possible which avoids
   unnecessary computation.

Revision: 06.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to output sample list
        - pointer to samples name
        - pointer to samples comment
        - preview
Output: - name updated as flagged
        - comment updated as flagged
        - output sample list updated as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_compose_sigin(double_lst src_fctn, int src_type, double_lst *fsum, 
		    char **tname, char **comment, int preview)
{
  SRC_IS source_available = SRC_INVALID;

  assert(tname);
  assert(comment);
  assert(fsum);

  if (src_type == DC_BIT_POS_PAD) {                   /* source dependence */
    if (!src_fctn.cnt) {                             /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else {
      source_available = SRC_VALID_COMPLETE;           /* source available */

      if (!preview) {
	add_string(DC_OPT_POS_PAD_FILE, tname);
	cat_string(DC_PREFIX, DC_OPT_POS_PAD_DESC, DC_SUFFIX, comment);

	assert(!fsum->cnt && !fsum->dbl);

	copy_double_lst(&src_fctn, fsum);
      }
    }
  }

  return source_available;
}



/* Dumps samples. A preview of source completion is possible which avoids
   unnecessary computation.

Revision: 06.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to output sample list
        - pointer to samples name
        - pointer to samples comment
        - preview
Output: - name updated as flagged
        - output sample list updated as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_compose_pos(double_lst src_fctn, int src_type, double_lst *fsum, 
		  char **tname, char **comment, int preview)
{
  SRC_IS source_available = SRC_INVALID;
  char   *stemp = 0;

  assert(tname);
  assert(comment);
  assert(fsum);

  if (src_type == DC_BIT_POS) {                       /* source dependence */
    if (!src_fctn.cnt) {                             /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else {
      source_available = SRC_VALID_COMPLETE;           /* source available */

      if (!preview) {
	add_string(DC_OPT_POS_FILE, tname);
	cat_string(POS_HDR, "\n", "", &stemp);
	cat_string(DC_PREFIX, DC_OPT_POS_DESC, DC_SUFFIX, &stemp);

	assert(!fsum->cnt && !fsum->dbl);

	copy_double_lst(&src_fctn, fsum);
      }
    }
  }

  FREEMEM(stemp);

  return source_available;
}



/* Dumps samples. A preview of source completion is possible which avoids
   unnecessary computation.

Revision: 16.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to output sample list
        - pointer to samples name
        - pointer to samples comment
        - preview
Output: - name updated as flagged
        - output sample list updated as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_compose_hphase(double_lst src_fctn, int src_type, double_lst *fsum, 
		     char **tname, char **comment, int preview)
{
  SRC_IS source_available = SRC_INVALID;
  char   *stemp = 0;

  assert(tname);
  assert(comment);
  assert(fsum);

  if (src_type == DC_BIT_POS_HPHASE) {                /* source dependence */
    if (!src_fctn.cnt) {                             /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else {
      source_available = SRC_VALID_COMPLETE;           /* source available */

      if (!preview) {
	add_string(DC_OPT_POS_HPHASE_FILE, tname);
	cat_string(DC_PREFIX, DC_OPT_POS_HPHASE_DESC, DC_SUFFIX, &stemp);

	assert(!fsum->cnt && !fsum->dbl);

	copy_double_lst(&src_fctn, fsum);
      }
    }
  }

  FREEMEM(stemp);

  return source_available;
}



/* Dumps samples. A preview of source completion is possible which avoids
   unnecessary computation.

Revision: 07.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to output sample list
        - pointer to samples name
        - pointer to samples comment
        - preview
Output: - name updated as flagged
        - comment updated as flagged
        - output sample list updated as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_compose_tfin(double_lst src_fctn, int src_type, double_lst *fsum, 
		  char **tname, char **comment, int preview)
{
  SRC_IS source_available = SRC_INVALID;

  assert(tname);
  assert(comment);
  assert(fsum);

  if (src_type == DC_BIT_TFIN) {                      /* source dependence */
    if (!src_fctn.cnt) {                             /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else {
      source_available = SRC_VALID_COMPLETE;           /* source available */

      if (!preview) {
	add_string(DC_OPT_TFIN_FILE, tname);
	cat_string(DC_PREFIX, DC_OPT_TFIN_DESC, DC_SUFFIX, comment);

	assert(!fsum->cnt && !fsum->dbl);

	copy_double_lst(&src_fctn, fsum);
      }
    }
  }

  return source_available;
}



/* Dumps samples. A preview of source completion is possible which avoids
   unnecessary computation.

Revision: 07.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to output sample list
        - pointer to samples name
        - pointer to samples comment
        - preview
Output: - name updated as flagged
        - comment updated as flagged
        - output sample list updated as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_compose_tflag(double_lst src_fctn, int src_type, double_lst *fsum, 
		    char **tname, char **comment, int preview)
{
  SRC_IS source_available = SRC_INVALID;

  assert(tname);
  assert(comment);
  assert(fsum);

  if (src_type == DC_BIT_TFLAG) {                     /* source dependence */
    if (!src_fctn.cnt) {                             /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else {
      source_available = SRC_VALID_COMPLETE;           /* source available */

      if (!preview) {
	add_string(DC_OPT_TFLAG_FILE, tname);
	cat_string(DC_PREFIX, DC_OPT_TFLAG_DESC, DC_SUFFIX, comment);

	assert(!fsum->cnt && !fsum->dbl);

	copy_double_lst(&src_fctn, fsum);
      }
    }
  }

  return source_available;
}



/* Builds grey document image source. A preview of source completion is 
   possible which avoids unnecessary computation.

Revision: 30.12.2002 E.Sack

Input : - source image
        - source image type
        - image list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_gv(pic_def src_pic, int src_type, dc_pic_lst dpl, pic_def *psum,
	       char **tname, int preview)
{
  int     idx_src = -1;
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic;

  assert(tname);

  init_pic_def(&pic);

  if (src_type != DC_BIT_GV) {                        /* source dependence */
    idx_src = dc_exist_source(dpl, DC_BIT_GV);              /* check stock */

    if (idx_src < 0) {                               /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else if (!preview) {
      pic = dc_load_source(dpl, idx_src);
    }
  }
  else if (!preview) {                              /* direct input source */
    copy_pic_def(&src_pic, &pic);
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    dc_overlay(pic, psum, foreground_color);

    add_string(DC_OPT_GV_FILE, tname);
  }

  free_pic_def(&pic);

  return source_available;
}



/* Builds purged binary document image source. A preview of source
   completion is possible which avoids unnecessary computation.

Revision: 30.12.2002 E.Sack

Input : - source image
        - source image type
        - image list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_clean(pic_def src_pic, int src_type, dc_pic_lst dpl, pic_def *psum,
                 char **tname, int preview)
{
  int     idx_src = -1;
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic;

  assert(tname);

  init_pic_def(&pic);

  if (src_type != DC_BIT_CLEAN) {                     /* source dependence */
    idx_src = dc_exist_source(dpl, DC_BIT_CLEAN);           /* check stock */

    if (idx_src < 0) {                               /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else if (!preview) { 
      pic = dc_load_source(dpl, idx_src);
    }
  }
  else if (!preview) {                              /* direct input source */
    copy_pic_def(&src_pic, &pic);
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    dc_overlay(pic, psum, foreground_color);

    add_string(DC_OPT_CLEAN_FILE, tname);
  }

  free_pic_def(&pic);

  return source_available;
}



/* Builds binary document image source. A preview of source
   completion is possible which avoids unnecessary computation.

Revision: 30.12.2002 E.Sack

Input : - source image
        - source image type
        - image list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_org(pic_def src_pic, int src_type, dc_pic_lst dpl, pic_def *psum,
	       char **tname, int preview)
{
  int     idx_src = -1;
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic;

  assert(tname);

  init_pic_def(&pic);

  if (src_type != DC_BIT_BMP) {                       /* source dependence */
    idx_src = dc_exist_source(dpl, DC_BIT_BMP);             /* check stock */

    if (idx_src < 0) {                               /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else if (!preview) {
      pic = dc_load_source(dpl, idx_src);
    }
  }
  else if (!preview) {                              /* direct input source */
    copy_pic_def(&src_pic, &pic);
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    dc_overlay(pic, psum, foreground_color);

    add_string(DC_OPT_BMP_FILE, tname);
  }

  free_pic_def(&pic);

  return source_available;
}



/* Builds RB framed overlays. A preview of source completion is possible
   which avoids unnecessary computation.

Revision: 30.12.2002 E.Sack

Input : - source RB list
        - pointer to overlay image
        - source sign
        - pointer to frame color (may be NULL)
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_frame(rb_lst src_rb, pic_def *psum, char *ext, 
		 col_def *frame_color, char **tname, int preview)
{
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic_frame;

  assert(tname);

  init_pic_def(&pic_frame);

  if (!src_rb.cnt) {
    source_available = SRC_VALID_INCOMPLETE;
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    if (psum->ar.width || psum->ar.height) {         /* ensure consistency */
      assert(psum->ar.width == src_rb.rar.width);
      assert(psum->ar.height == src_rb.rar.height);
    }

    pic_frame = dc_copy_frames(src_rb, frame_color);

    dc_overlay(pic_frame, psum, foreground_color);

    add_string(ext, tname);
  }

  free_pic_def(&pic_frame);

  return source_available;
}



/* Builds horizontal lines. A preview of source completion is possible
   which avoids unnecessary computation.

Revision: 04.01.2003 E.Sack

Input : - source horizontal lines list
        - image list
        - pointer to overlay image
        - source sign
        - pointer to line color (may be NULL)
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_hline(int_lst hline, dc_pic_lst dpl, pic_def *psum, char *ext, 
		 col_def *line_color, char **tname, int preview)
{
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic_hline;

  assert(tname);

  init_pic_def(&pic_hline);

  if (!hline.cnt || !AREA_SIZE(dpl.sar)) {
    source_available = SRC_VALID_INCOMPLETE;
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    pic_hline = dc_copy_lines(dpl.sar, line_color, hline);

    dc_overlay(pic_hline, psum, foreground_color);

    add_string(ext, tname);
  }

  free_pic_def(&pic_hline);

  return source_available;
}



/* Builds horizontal subdivision lines of multiline RBs. A preview of source
   completion is possible which avoids unnecessary computation.

Revision: 05.01.2003 E.Sack

Input : - source RB list
        - pointer to overlay image
        - source sign
        - pointer to line color (may be NULL)
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_sline(rb_lst src_rb, pic_def *psum, char *ext, 
		 col_def *sline_color, char **tname, int preview)
{
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic_sline;

  assert(tname);

  init_pic_def(&pic_sline);

  if (!src_rb.cnt) {
    source_available = SRC_VALID_INCOMPLETE;
  }
  else {
    assert(AREA_SIZE(src_rb.rar));
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    pic_sline = dc_copy_grid(src_rb, sline_color);

    dc_overlay(pic_sline, psum, foreground_color);

    add_string(ext, tname);
  }

  free_pic_def(&pic_sline);

  return source_available;
}



/* Builds image of failed RB recognition. A preview of source completion is 
   possible which avoids unnecessary computation.

Revision: 30.12.2002 E.Sack

Input : - source candidate list
        - pattern type flag
        - method character
        - source character
        - image list
        - source flag
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_rb(cand_lst hlist, PATTERN_ATT pattern_flag, char *ext_method, 
	      char *ext_src, dc_pic_lst dpl, int src_flag, 
	      PIXEL_KEY grey_backgnd, pic_def *psum, char **tname, 
	      int preview)
{
  int     idx_src = -1;
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic_src, pic_fail;

  assert(tname);

  init_pic_def(&pic_src);
  init_pic_def(&pic_fail);

  if (!src_flag) {
    source_available = SRC_INVALID;
    return source_available;
  }

  if (!hlist.cnt) {
    source_available = SRC_VALID_INCOMPLETE;
  }
  else {
    idx_src = dc_exist_source(dpl, src_flag);               /* check stock */

    if (idx_src < 0) {                               /* check availability */
      source_available = SRC_VALID_INCOMPLETE;     /* source not available */
    }
    else if (!preview) {
      pic_src = dc_load_source(dpl, idx_src);
    }
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    pic_fail = dc_cinderella_pattern(pic_src, hlist, pattern_flag, 
				     grey_backgnd);

    dc_overlay(pic_fail, psum, foreground_color);

    cat_string(ext_src, ext_method, "", tname);
  }

  free_pic_def(&pic_src);
  free_pic_def(&pic_fail);

  return source_available;
}



/* Builds image of 2-D double numbers. A preview of source completion is 
   possible which avoids unnecessary computation.

Revision: 10.01.2003 E.Sack

Input : - source image
        - source character
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_ccf(pic_dbl dpic, char *ext_src, pic_def *psum, char **tname, 
	       int preview)
{
  SRC_IS source_available = SRC_VALID_COMPLETE;

  assert(tname);

  if (!dpic.mem) {
    source_available = SRC_VALID_INCOMPLETE;
  }
  else {
    assert(AREA_SIZE(dpic.ar));
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    assert(!psum->mem);

    *psum = convert_pic_dbl(dpic);

    add_string(ext_src, tname);
  }

  return source_available;
}



/* Builds image of recognized NB. A preview of source completion is 
   possible which avoids unnecessary computation.

Revision: 31.12.2002 E.Sack

Input : - source candidate list
        - pointer to source classes (may be discarded)
        - image list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - name updated as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_nb(cand_lst hlist, cls_def *cls, dc_pic_lst dpl, pic_def *psum, 
	      char **tname, int preview)
{
  SRC_IS  source_available = SRC_VALID_COMPLETE;
  pic_def pic_nb;

  assert(tname);

  init_pic_def(&pic_nb);

  if (!hlist.cnt || !cls || !AREA_SIZE(dpl.sar)) {
    source_available = SRC_VALID_INCOMPLETE;
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    col_def foreground_color = FGND_COLOR;

    pic_nb = dc_cinderella_nb(dpl.sar, hlist, cls);

    dc_overlay(pic_nb, psum, foreground_color);

    add_string(DC_OPT_NB_GOOD_FILE, tname);
  }

  free_pic_def(&pic_nb);

  return source_available;
}



/* Wrapper for all online sources.

Revision: 31.12.2002 E.Sack

Input : - RB source
        - MRB source
        - candidate list source
        - pointer to classes
        - horizontal line list source
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_all(rb_lst src_rb, rb_lst src_mrb, cand_lst src_cand, 
	       cls_def *src_cls, int_lst src_hline, char *fname, 
	       dc_pic_lst *dpl, int *flag_list, int preview)
{
  int     dummy_type = DC_BIT_EMPTY;
  SRC_IS  src_available;
  pic_def dummy_pic;

  assert(src_cls);
  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(flag_list);

  init_pic_def(&dummy_pic);

  src_available = dc_composition(dummy_pic, dummy_type, src_rb, src_mrb,
				 src_cand, src_cls, src_hline, fname, 
				 dpl, flag_list, preview);
  return src_available;
}



/* Wrapper for image source.

Revision: 30.12.2002 E.Sack

Input : - image source
        - source type
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_pic(pic_def src_pic, int src_type, char *fname, dc_pic_lst *dpl, 
	       int *flag_list, int preview)
{
  SRC_IS   src_available;
  rb_lst   dummy_rb;
  rb_lst   dummy_mrb;
  cand_lst dummy_cand;
  cls_def  *dummy_cls = 0;
  int_lst  dummy_hline;

  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(flag_list);

  init_rb_lst(&dummy_rb);
  init_rb_lst(&dummy_mrb);
  init_cand_lst(&dummy_cand);
  init_int_lst(&dummy_hline);

  src_available = dc_composition(src_pic, src_type, dummy_rb, dummy_mrb,
				 dummy_cand, dummy_cls, dummy_hline, 
				 fname, dpl, flag_list, preview);
  return src_available;
}



/* Wrapper for image source.

Revision: 30.12.2002 E.Sack

Input : - source image
        - source type
        - pointer to filename
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_direct_pic_dbl(pic_dbl src_dpic, int src_type, char *fname, 
		  int *flag_list, int preview)
{
  SRC_IS   src_available;

  assert(fname);
  assert(strlen(fname));
  assert(flag_list);

  src_available = dc_direct(src_dpic, src_type, fname, flag_list, preview);

  return src_available;
}



/* Wrapper for RB list source.

Revision: 30.12.2002 E.Sack

Input : - RB source
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_rb(rb_lst src_rb, char *fname, dc_pic_lst *dpl, int *flag_list, 
	      int preview)
{
  int      dummy_type = DC_BIT_EMPTY;
  SRC_IS   src_available;
  pic_def  dummy_pic;
  cand_lst dummy_cand;
  cls_def *dummy_cls = 0;
  rb_lst   dummy_mrb;
  int_lst  dummy_hline;

  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(flag_list);

  init_pic_def(&dummy_pic);
  init_cand_lst(&dummy_cand);
  init_rb_lst(&dummy_mrb);
  init_int_lst(&dummy_hline);

  src_available = dc_composition(dummy_pic, dummy_type, src_rb, 
				 dummy_mrb, dummy_cand, dummy_cls, 
				 dummy_hline, fname, dpl, flag_list, 
				 preview);
  return src_available;
}



/* Wrapper for RB list source.

Revision: 30.12.2002 E.Sack

Input : - MRB source
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_mrb(rb_lst src_mrb, char *fname, dc_pic_lst *dpl, int *flag_list,
	       int preview)
{
  int      dummy_type = DC_BIT_EMPTY;
  SRC_IS   src_available;
  pic_def  dummy_pic;
  cand_lst dummy_cand;
  cls_def *dummy_cls = 0;
  rb_lst   dummy_rb;
  int_lst  dummy_hline;

  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(flag_list);

  init_pic_def(&dummy_pic);
  init_cand_lst(&dummy_cand);
  init_rb_lst(&dummy_rb);
  init_int_lst(&dummy_hline);

  src_available = dc_composition(dummy_pic, dummy_type, dummy_rb, 
				 src_mrb, dummy_cand, dummy_cls, 
				 dummy_hline, fname, dpl, flag_list, 
				 preview);
  return src_available;
}



/* Wrapper for candidate list source.

Revision: 30.12.2002 E.Sack

Input : - candidate list source
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_pattern(cand_lst src_cand, char *fname, dc_pic_lst *dpl, 
		   int *flag_list, int preview)
{
  int     dummy_type = DC_BIT_EMPTY;
  SRC_IS  src_available;
  pic_def dummy_pic;
  rb_lst  dummy_rb;
  rb_lst  dummy_mrb;
  cls_def *dummy_cls = 0;
  int_lst dummy_hline;

  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(flag_list);

  init_pic_def(&dummy_pic);
  init_rb_lst(&dummy_rb);
  init_rb_lst(&dummy_mrb);
  init_int_lst(&dummy_hline);

  src_available = dc_composition(dummy_pic, dummy_type, dummy_rb, 
				 dummy_mrb, src_cand, dummy_cls, 
				 dummy_hline, fname, dpl, flag_list, 
				 preview);
  return src_available;
}



/* Wrapper for class list source.

Revision: 31.12.2002 E.Sack

Input : - candidate list source
        - pointer to classes
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_nb(cand_lst src_cand, cls_def *src_cls, char *fname, 
	      dc_pic_lst *dpl, int *flag_list, int preview)
{
  int     dummy_type = DC_BIT_EMPTY;
  SRC_IS  src_available;
  pic_def dummy_pic;
  rb_lst  dummy_rb;
  rb_lst  dummy_mrb;
  int_lst dummy_hline;

  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(src_cls);
  assert(flag_list);

  init_pic_def(&dummy_pic);
  init_rb_lst(&dummy_rb);
  init_rb_lst(&dummy_mrb);
  init_int_lst(&dummy_hline);

  src_available = dc_composition(dummy_pic, dummy_type, dummy_rb, 
				 dummy_mrb, src_cand, src_cls, 
				 dummy_hline, fname, dpl, flag_list, 
				 preview);
  return src_available;
}



/* Wrapper for class list source.

Revision: 31.12.2002 E.Sack

Input : - horizontal line list source
        - pointer to classes
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_hline(int_lst src_hline, char *fname, dc_pic_lst *dpl, 
		 int *flag_list, int preview)
{
  int      dummy_type = DC_BIT_EMPTY;
  SRC_IS   src_available;
  pic_def  dummy_pic;
  rb_lst   dummy_rb;
  rb_lst   dummy_mrb;
  cand_lst dummy_cand;
  cls_def  *dummy_cls = 0;

  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(flag_list);

  init_pic_def(&dummy_pic);
  init_rb_lst(&dummy_rb);
  init_rb_lst(&dummy_mrb);
  init_cand_lst(&dummy_cand);

  src_available = dc_composition(dummy_pic, dummy_type, dummy_rb, 
				 dummy_mrb, dummy_cand, dummy_cls, 
				 src_hline, fname, dpl, flag_list, preview);
  return src_available;
}



/* Wrapper for FFT source.

Revision: 06.01.2003 E.Sack

Input : - source FFT
        - source type
        - pointer to filename
        - pointer to subject string
        - pointer to flag list
        - preview
Output: - FFT stored as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_ccf(fft_def src_fft, int src_type, char *fname, char *subject, 
	  int *flag_list, int preview)
{
  SRC_IS     src_available;
  double_lst dummy_fctn;

  assert(fname);
  assert(strlen(fname));
  assert(flag_list);
  assert(subject);

  init_double_lst(&dummy_fctn);

  src_available = dc_1d_composition(dummy_fctn, src_fft, src_type, fname, 
				    subject, flag_list, preview);
  return src_available;
}



/* Wrapper for sample list source.

Revision: 06.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to filename
        - pointer to subject string
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_double(double_lst src_fctn, int src_type, char *fname, char *subject,
	     int *flag_list, int preview)
{
  SRC_IS  src_available;
  fft_def dummy_fft;

  assert(fname);
  assert(strlen(fname));
  assert(flag_list);
  assert(subject);

  init_fft_def(&dummy_fft);

  src_available = dc_1d_composition(src_fctn, dummy_fft, src_type, fname, 
				    subject, flag_list, preview);
  return src_available;
}



/* Dumps all flagged sources of the documentation channel, if they are
   complete, direct to disk. A preview of source completion is possible
   which avoids unnecessary computation.

Revision: 10.01.2003 E.Sack

Input : - image source
        - source type
        - pointer to filename
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_direct(pic_dbl src_dpic, int src_type, char *fname, int *flag_list, 
	  int preview)
{
  SRC_IS  source_available = SRC_INVALID;
  pic_def psum;
  char    *tname = 0;

  assert(fname);
  assert(strlen(fname));
  assert(flag_list);

  init_pic_def(&psum);

  tname = dc_base_name(fname, DC_EXT_2D);

  while (1) {
    if ((*flag_list & DC_BIT_PEST) && (DC_BIT_PEST == src_type)) {
      source_available = dc_compose_ccf(src_dpic, DC_OPT_PEST_FILE, &psum, 
					&tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }

      if (source_available == SRC_VALID_COMPLETE) {
	break;                                    /* no further processing */
      }
    }

    if ((*flag_list & DC_BIT_PEST_MAX) && (DC_BIT_PEST_MAX == src_type)) {
      source_available = dc_compose_ccf(src_dpic, DC_OPT_PEST_MAX_FILE, 
					&psum, &tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }

      if (source_available == SRC_VALID_COMPLETE) {
	break;                                    /* no further processing */
      }
    }

    break;
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    dump_pic_def(psum, tname, psum.ar, 0, 0, xctrl.cfg.verbose > BE_QUIET,
		 PNM_FORMAT_UNDEF);
    *flag_list = DC_BIT_EMPTY;            /* option processed successfully */
  }

  FREEMEM(tname);
  free_pic_def(&psum);

  return source_available;
}



  /* Overlays and stores all flagged images of the documentation channel
   if sources are complete. A preview of source completion is possible
   which avoids unnecessary computation. Completeness of needed sources as
   forward dependence.

Revision: 28.12.2002 E.Sack

Input : - image source
        - source type
        - RB source
        - MRB source
        - candidate list source
        - pointer to classes (may be discarded)
        - horizontal line list source
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - preview
Output: - image stored as flagged
        - source availability flag
Implementation: As only image sources are stored temporarily in the DC,
                other sources than images must be offered to the
                documentation channel online. Consequently this function has
                a big interface and the function should be called from 
                wrappers. If all data are available simultaneously, this
                function should be called directly. This is supposed at 
                least once.
*/
static SRC_IS
dc_composition(pic_def src_pic, int src_type, rb_lst src_rb, 
	       rb_lst src_mrb, cand_lst src_cand, cls_def *src_cls, 
	       int_lst src_hline, char *fname, dc_pic_lst *dpl, 
	       int *flag_list, int preview)
{
  int     old_flag_list = *flag_list;
  SRC_IS  source_available;
  pic_def psum;
  char    *tname = 0;

  assert(fname);
  assert(strlen(fname));
  assert(dpl);
  assert(flag_list);

  init_pic_def(&psum);

  tname = dc_base_name(fname, DC_EXT_2D);

  source_available = dc_overlay_method(src_cand, *dpl, &old_flag_list, 
				       &psum, &tname, preview);

  if (source_available == SRC_VALID_INCOMPLETE) {       /* no more overlay */
    assert(!psum.mem && !psum.col);
  }
  else if (source_available != SRC_VALID_COMPLETE) {       /* more overlay */
    source_available = dc_compose_overlay(src_pic, src_type, src_rb, src_mrb, 
					  src_cand, src_cls, src_hline, 
					  fname, *dpl, &old_flag_list, &psum,
					  &tname, preview);
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    dc_update(psum, dpl, *flag_list, tname);

    *flag_list = old_flag_list;           /* option processed successfully */
  }

  FREEMEM(tname);
  free_pic_def(&psum);

  return source_available;
}



/* Updates documentation channel store.

Revision: 10.01.2003 E.Sack

Input : - overlay image
        - pointer to image source list
        - pointer to flag list
        - pointer to filename
Output: - image source list updated
*/
static void
dc_update(pic_def psum, dc_pic_lst *dpl, int flag_list, char *tname)
{
  if (psum.mem || psum.col) { /* there may be options without consequences */
    int src_count;

    dump_pic_def(psum, tname, psum.ar, 0, 0, xctrl.cfg.verbose > BE_QUIET,
		 PNM_FORMAT_UNDEF);

    src_count = dpl->cnt++;

    if (src_count) {
      REALLOC(dc_pic_src, dpl->cnt, dpl->dps);
    }
    else {
      CALLOC(dc_pic_src, dpl->cnt, dpl->dps);
    }

    init_dc_pic_src(&dpl->dps[src_count]);

    add_string(tname, &dpl->dps[src_count].tname);
    dpl->dps[src_count].req = flag_list;             /* stored composition */
  }
}



/* Dumps sampled functions of the documentation channel.

Revision: 06.01.2003 E.Sack

Input : - source sample list
        - source FFT
        - source type
        - pointer to filename
        - pointer to subject string
        - pointer to flag list
        - preview
Output: - sample list stored as flagged
        - source availability flag
*/
static SRC_IS
dc_1d_composition(double_lst src_fctn, fft_def src_fft, int source_type,
		  char *fname, char *subject, int *flag_list, int preview)
{
  int        old_flag_list = *flag_list;
  SRC_IS     source_available;
  char       *tname = 0;
  char       *comment = 0;
  double_lst fsum;

  assert(fname);
  assert(strlen(fname));
  assert(flag_list);
  assert(subject);

  init_double_lst(&fsum);

  tname = dc_base_name(fname, DC_EXT_1D);

  source_available = dc_1d_compose(src_fctn, src_fft, source_type, fname, 
				   &old_flag_list, &fsum, &tname, &comment, 
				   preview);

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    assert(fsum.dbl);
    assert(fsum.cnt);
    assert(tname);

    dump_double(fsum.dbl, fsum.cnt, tname, 0, comment, subject);

    *flag_list = old_flag_list;           /* option processed successfully */
  }

  free_double_lst(&fsum);
  FREEMEM(tname);
  FREEMEM(comment);

  return source_available;
}



/* Performs different methods for overlays of the documentation channel
   if sources are complete. A preview of source completion is possible
   which avoids unnecessary computation. 

Revision: 04.01.2003 E.Sack

Input : - source candidate list
        - pattern type flag
        - method character
        - image list
        - pointer to flag list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_method_rb(cand_lst src_cand, PATTERN_ATT pattern_flag, 
		     char *ext_method, dc_pic_lst dpl, int flag_list, 
		     PIXEL_KEY grey_backgnd, pic_def *psum, char **tname, 
		     int preview)
{
  SRC_IS source_available, src_tmp_available;

  assert(tname);
  assert(flag_list);

  while(1) {
    source_available = dc_compose_rb(src_cand, pattern_flag, ext_method, 
				     DC_OPT_ORG_FILE, dpl, 
				     flag_list & DC_BIT_ORG, 
				     grey_backgnd, psum, tname, preview);
    if (source_available == SRC_VALID_INCOMPLETE) {
      break;
    }

    src_tmp_available = dc_compose_rb(src_cand, pattern_flag, ext_method, 
				      DC_OPT_BMP_FILE, dpl, 
				      flag_list & DC_BIT_BMP, 
				      grey_backgnd, psum, tname, preview);
    if (src_tmp_available == SRC_VALID_INCOMPLETE) {
      source_available = src_tmp_available;
      break;                         /* first incomplete composition stops */
    }
    else if (source_available != SRC_VALID_COMPLETE) {
      source_available = src_tmp_available;
    }

    src_tmp_available = dc_compose_rb(src_cand, pattern_flag, ext_method, 
				      DC_OPT_CLEAN_FILE, dpl,
				      flag_list & DC_BIT_CLEAN, 
				      grey_backgnd, psum, tname, preview);

    if (src_tmp_available == SRC_VALID_INCOMPLETE) {
      source_available = src_tmp_available;
      break;                         /* first incomplete composition stops */
    }
    else if (source_available != SRC_VALID_COMPLETE) {
      source_available = src_tmp_available;
    }

    src_tmp_available = dc_compose_rb(src_cand, pattern_flag, ext_method,
				      DC_OPT_GV_FILE, dpl,
				      flag_list & DC_BIT_GV, 
				      grey_backgnd, psum, tname, preview);

    if (src_tmp_available == SRC_VALID_INCOMPLETE) {
      source_available = src_tmp_available;
      break;                         /* first incomplete composition stops */
    }
    else if (source_available != SRC_VALID_COMPLETE) {
      source_available = src_tmp_available;
    }

    break;
  }

  return source_available;
}



/* Performs different methods for overlays of the documentation channel
   if sources are complete. A preview of source completion is possible
   which avoids unnecessary computation. 

Revision: 04.01.2003 E.Sack

Input : - source candidate list
        - pointer to classes (may be discarded)
        - image list
        - pointer to flag list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - overlay image updated as flagged
        - source availability flag
*/
static SRC_IS
dc_overlay_method(cand_lst src_cand, dc_pic_lst dpl, 
		  int *flag_list, pic_def *psum, char **tname, int preview)
{
  SRC_IS source_available = SRC_INVALID;

  assert(tname);
  assert(flag_list);

  while (1) {
    if (*flag_list & DC_BIT_RB_FAIL) {
      source_available = dc_overlay_method_rb(src_cand, PATTERN_UREC,
					      DC_OPT_RB_FAIL_FILE, dpl, 
					      *flag_list, BACKGROUND, 
					      psum, tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }
    }

    if (*flag_list & DC_BIT_RB_FAIL_INV) {
      source_available = dc_overlay_method_rb(src_cand, PATTERN_UREC,
					      DC_OPT_RB_FAIL_INV_FILE, dpl, 
					      *flag_list, FOREGROUND, 
					      psum, tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }
    }

    if (*flag_list & DC_BIT_RB_GOOD) {
      source_available = dc_overlay_method_rb(src_cand, PATTERN_REC,
					      DC_OPT_RB_GOOD_FILE, dpl, 
					      *flag_list, BACKGROUND, 
					      psum, tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }
    }

    if (*flag_list & DC_BIT_RB_GOOD_INV) {
      source_available = dc_overlay_method_rb(src_cand, PATTERN_REC,
					      DC_OPT_RB_GOOD_INV_FILE, dpl, 
					      *flag_list, FOREGROUND, 
					      psum, tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }
    }

    if (*flag_list & DC_BIT_RB_ALL) {
      source_available = dc_overlay_method_rb(src_cand, PATTERN_ALL,
					      DC_OPT_RB_ALL_FILE, dpl, 
					      *flag_list, BACKGROUND, 
					      psum, tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }
    }

    if (*flag_list & DC_BIT_RB_ALL_INV) {
      source_available = dc_overlay_method_rb(src_cand, PATTERN_ALL,
					      DC_OPT_RB_ALL_INV_FILE, dpl, 
					      *flag_list, FOREGROUND, 
					      psum, tname, preview);
      if (source_available == SRC_VALID_INCOMPLETE) {
	break;                       /* first incomplete composition stops */
      }
    }

    break;
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    *flag_list = DC_BIT_EMPTY;
  }

  return source_available;
}



/* Overlays all flagged images of the documentation channel if sources
   are complete. A preview of source completion is possible which avoids
   unnecessary computation.

Revision: 19.01.2003 E.Sack

Input : - image source
        - source type
        - RB source
        - MRB source
        - horizontal line list source
        - pointer to filename
        - pointer to image source list
        - pointer to flag list
        - pointer to overlay image
        - pointer to overlay name
        - preview
Output: - image stored as flagged
        - source availability flag
*/
static SRC_IS
dc_compose_overlay(pic_def src_pic, int src_type, rb_lst src_rb, 
		   rb_lst src_mrb, cand_lst src_cand, cls_def *src_cls, 
		   int_lst src_hline, char *fname, dc_pic_lst dpl, 
		   int *flag_list, pic_def *psum, char **tname, int preview)
{
  int    bit, flag, old_flag_list;
  SRC_IS source_available = SRC_INVALID;

  assert(tname);
  assert(fname);
  assert(strlen(fname));
  assert(flag_list);

  old_flag_list = *flag_list;

  for (bit = 0; bit < 32; bit++) {                       /* test flag list */
    flag = old_flag_list & (1<<bit);

    switch (flag) {          /* sampling overlays: ordains file extensions */
    case DC_BIT_EMPTY:                                       /* empty flag */
    case DC_BIT_PEST:                                        /* X-Y-method */
    case DC_BIT_PEST_MAX:                                    /* X-Y-method */
    case DC_BIT_RB_FAIL:                                   /* other method */
    case DC_BIT_RB_FAIL_INV:                               /* other method */
    case DC_BIT_RB_GOOD:                                   /* other method */
    case DC_BIT_RB_GOOD_INV:                               /* other method */
    case DC_BIT_RB_ALL:                                    /* other method */
    case DC_BIT_RB_ALL_INV:                                /* other method */
      break;
    case DC_BIT_NB_GOOD:
      source_available = dc_compose_nb(src_cand, src_cls, dpl, psum, 
				       tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_NB_GOOD;
      }
      break;
    case DC_BIT_ORG:
      source_available = dc_compose_doc(src_pic, src_type, dpl, psum, 
					tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_ORG;
      }
      break;
    case DC_BIT_BMP:
      source_available = dc_compose_org(src_pic, src_type, dpl, psum, 
					tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_BMP;
      }
      break;
    case DC_BIT_CLEAN:
      source_available = dc_compose_clean(src_pic, src_type, dpl, psum, 
				          tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_CLEAN;
      }
      break;
    case DC_BIT_DOT:
      source_available = dc_compose_dot(src_pic, src_type, dpl, psum, 
					tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_DOT;
      }
      break;
    case DC_BIT_MRB_FRAME_BMP:
      source_available = dc_compose_frame(src_mrb, psum,
					  DC_OPT_MRB_FRAME_BMP_FILE, 0,
					  tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_MRB_FRAME_BMP;
      }
      break;
    case DC_BIT_MRB_FRAME_RGB:
      {
	col_def frame_color = MRB_FRAME_COLOR;

	source_available = dc_compose_frame(src_mrb, psum,
					    DC_OPT_MRB_FRAME_RGB_FILE, 
					    &frame_color, tname, preview);
	if (source_available == SRC_VALID_COMPLETE) {
	  old_flag_list &= ~DC_BIT_MRB_FRAME_RGB;
	}
      }
      break;
    case DC_BIT_RB_FRAME_BMP:
	source_available = dc_compose_frame(src_rb, psum,
					    DC_OPT_RB_FRAME_BMP_FILE, 0,
					    tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_RB_FRAME_BMP;
      }
      break;
    case DC_BIT_RB_FRAME_RGB:
      {
	col_def frame_color = RB_FRAME_COLOR;

	source_available = dc_compose_frame(src_rb, psum,
					    DC_OPT_RB_FRAME_RGB_FILE, 
					    &frame_color, tname, preview);
	if (source_available == SRC_VALID_COMPLETE) {
	  old_flag_list &= ~DC_BIT_RB_FRAME_RGB;
	}
      }
      break;
    case DC_BIT_HLINE_BMP:
      source_available = dc_compose_hline(src_hline, dpl, psum,
					  DC_OPT_HLINE_BMP_FILE, 0,
					  tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_HLINE_BMP;
      }
      break;
    case DC_BIT_HLINE_RGB:
      {
	col_def line_color = HLINE_COLOR;

	source_available = dc_compose_hline(src_hline, dpl, psum, 
					    DC_OPT_HLINE_RGB_FILE, 
					    &line_color, tname, preview);
	if (source_available == SRC_VALID_COMPLETE) {
	  old_flag_list &= ~DC_BIT_HLINE_RGB;
	}
      }
      break;
    case DC_BIT_SLINE_BMP:
      source_available = dc_compose_sline(src_rb, psum,
					  DC_OPT_SLINE_BMP_FILE, 0,
					  tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_SLINE_BMP;
      }
      break;
    case DC_BIT_SLINE_RGB:
      {
	col_def sline_color = SLINE_COLOR;

	source_available = dc_compose_sline(src_rb, psum,
					    DC_OPT_SLINE_RGB_FILE, 
					    &sline_color, tname, preview);
	if (source_available == SRC_VALID_COMPLETE) {
	  old_flag_list &= ~DC_BIT_SLINE_RGB;
	}
      }
      break;
    case DC_BIT_GV:
      source_available = dc_compose_gv(src_pic, src_type, dpl, psum, 
				       tname, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~DC_BIT_GV;
      }
      break;
    default:                                             /* no composition */
    fprintf(stderr, "\nERROR: composition missing for '0x%X'\n\n", flag);
      assert(0);
      break;
    }

    if (source_available == SRC_VALID_INCOMPLETE) {
      break;                         /* first incomplete composition stops */
    }
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    *flag_list = old_flag_list;
  }

  return source_available;
}



/* Dumps samples of the documentation channel if sources are complete. 
   A preview of source completion is possible which avoids unnecessary
   computation.

Revision: 06.01.2003 E.Sack

Input : - source sample list
        - source FFT
        - source type
        - pointer to filename
        - pointer to flag list
        - pointer to output sample list
        - pointer to sample name
        - pointer to comment string
        - preview
Output: - sample list stored as flagged
        - flag (0: not all sources available, no image stored
                1: all sources available, image stored)
*/
static SRC_IS
dc_1d_compose(double_lst src_fctn, fft_def src_fft, int src_type,
	      char *fname, int *flag_list, double_lst *fsum, char **tname, 
	      char **comment, int preview)
{
  int    bit, flag, old_flag_list;
  SRC_IS source_available = SRC_INVALID;

  assert(tname);
  assert(fname);
  assert(strlen(fname));
  assert(flag_list);
  assert(comment);

  old_flag_list = *flag_list;

  for (bit = 0; bit < 32; bit++) {                       /* test flag list */
    flag = old_flag_list & (1<<bit);

    switch (flag) {
    case DC_BIT_EMPTY:                                       /* empty flag */
      break;
    case DC_BIT_POS_CCF:
      source_available = dc_1d_compose_ccf(src_fft, src_type, fsum, tname,
					   comment, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~flag;
      }
      break;
    case DC_BIT_POS_PAD:
      source_available = dc_1d_compose_sigin(src_fctn, src_type, fsum, tname, 
					     comment, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~flag;
      }
      break;
    case DC_BIT_POS:
      source_available = dc_1d_compose_pos(src_fctn, src_type, fsum, tname, 
					   comment, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~flag;
      }
      break;
    case DC_BIT_POS_HPHASE:
      source_available = dc_1d_compose_hphase(src_fctn, src_type, fsum, 
					      tname, comment, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~flag;
      }
      break;
    case DC_BIT_TFIN:
      source_available = dc_1d_compose_tfin(src_fctn, src_type, fsum, tname, 
					    comment, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~flag;
      }
      break;
    case DC_BIT_TFLAG:
      source_available = dc_1d_compose_tflag(src_fctn, src_type, fsum, tname, 
					     comment, preview);
      if (source_available == SRC_VALID_COMPLETE) {
	old_flag_list &= ~flag;
      }
      break;
    default:                                             /* no composition */
      fprintf(stderr, "\nERROR: 1-D composition missing for '0x%X'\n\n", 
	      flag);
      assert(0);
      break;
    }

    if (source_available == SRC_VALID_COMPLETE) {
      break;                           /* first complete composition stops */
    }
  }

  if (!preview && source_available == SRC_VALID_COMPLETE) {
    *flag_list = old_flag_list;
  }

  return source_available;
}



/* Stores documentation channel composed images. If an overlay has all its
   components, it is stored to disk and its flags are cleared.

Revision: 28.12.2002 E.Sack

Input : - image source
        - source type
        - pointer to documentation channel
Output: - overlays stored
        - documentation channel updated
*/
static void
dc_store_pic(char *dc_name, pic_def src_pic, int src_type, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_overlay_pic(src_pic, src_type, dc_name,
				   &dcp->dpl, &option_test, 1);
    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_overlay_pic(src_pic, src_type, dc_name,
				   &dcp->dpl, &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    assert((option == option_test) || !option_test);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed images. If an overlay has all its
   components, it is stored to disk and its flags are cleared.

Revision: 27.03.2003 E.Sack

Input : - RB source
        - pointer to documentation channel
Output: - overlays stored
        - documentation channel updated
*/
static void
dc_store_rb(char *dc_name, rb_lst src_rb, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_overlay_rb(src_rb, dc_name, &dcp->dpl, 
				  &option_test, 1);
    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_overlay_rb(src_rb, dc_name, &dcp->dpl, 
				  &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed images. If an overlay has all its
   components, it is stored to disk and its flags are cleared.

Revision: 31.12.2002 E.Sack

Input : - MRB source
        - pointer to documentation channel
Output: - overlays stored
        - documentation channel updated
*/
static void
dc_store_mrb(char *dc_name, rb_lst src_mrb, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_overlay_mrb(src_mrb, dc_name, &dcp->dpl, 
				   &option_test, 1);
    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_overlay_mrb(src_mrb, dc_name, &dcp->dpl, 
				   &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    assert((option == option_test) || !option_test);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed images. If an overlay has all its
   components, it is stored to disk and its flags are cleared.

Revision: 31.12.2002 E.Sack

Input : - candidate list source
        - pointer to documentation channel
Output: - overlays stored
        - documentation channel updated
*/
static void
dc_store_pattern(char *dc_name, cand_lst src_cand, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_overlay_pattern(src_cand, dc_name, &dcp->dpl,
				       &option_test, 1);
    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_overlay_pattern(src_cand, dc_name, &dcp->dpl,
				       &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    assert((option == option_test) || !option_test);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed images. If an overlay has all its
   components, it is stored to disk and its flags are cleared.

Revision: 31.12.2002 E.Sack

Input : - candidate list source
        - pointer to classes
        - pointer to documentation channel
Output: - overlays stored
        - documentation channel updated
*/
static void
dc_store_nb(char *dc_name, cand_lst src_cand, cls_def *src_cls, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_overlay_nb(src_cand, src_cls, dc_name, 
				  &dcp->dpl, &option_test, 1);
    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_overlay_nb(src_cand, src_cls, dc_name, 
				  &dcp->dpl, &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    assert((option == option_test) || !option_test);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed images. If an overlay has all its
   components, it is stored to disk and its flags are cleared.

Revision: 31.12.2002 E.Sack

Input : - RB source
        - MRB source
        - candidate list source
        - pointer to classes
        - horizontal line list source
        - pointer to documentation channel
Output: - overlays stored
        - documentation channel updated
*/
static void
dc_store_all(char *dc_name, rb_lst src_rb, rb_lst src_mrb, cand_lst src_cand, 
	     cls_def *src_cls, int_lst src_hline, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_overlay_all(src_rb, src_mrb, src_cand, src_cls, 
				   src_hline, dc_name, &dcp->dpl, 
				   &option_test, 1);

    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_overlay_all(src_rb, src_mrb, src_cand, src_cls, 
				   src_hline, dc_name, &dcp->dpl, 
				   &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    assert((option == option_test) || !option_test);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed images. If an overlay has all its
   components, it is stored to disk and its flags are cleared.

Revision: 31.12.2002 E.Sack

Input : - horizontal line list source
        - pointer to documentation channel
Output: - overlays stored
        - documentation channel updated
*/
static void
dc_store_hline(char *dc_name, int_lst src_hline, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_overlay_hline(src_hline, dc_name, &dcp->dpl, 
				     &option_test, 1);
    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_overlay_hline(src_hline, dc_name, &dcp->dpl, 
				     &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    assert((option == option_test) || !option_test);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed samples. If a sample list has all
   its components, it is stored to disk and its flags are cleared.

Revision: 06.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to subject string
        - pointer to documentation channel
Output: - samples stored
        - documentation channel updated
*/
static void
dc_1d_store_double(char *dc_name, double_lst src_fctn, int source_type, 
		   char *subject, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);
  assert(subject);

  for (i = 0; i < dcp->opt_1d.cnt; i++) {                   /* option list */
    option = dcp->opt_1d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_1d_double(src_fctn, source_type, dc_name, subject,
				 &option_test, 1);
    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_1d_double(src_fctn, source_type, dc_name, subject,
				 &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    dcp->opt_1d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed samples. If a sample list has all
   its components, it is stored to disk and its flags are cleared.

Revision: 06.01.2003 E.Sack

Input : - source sample list
        - source type
        - pointer to subject string
        - pointer to documentation channel
Output: - samples stored
        - documentation channel updated
*/
static void
dc_store_pic_dbl(char *dc_name, pic_dbl source_dpic, int source_type, 
		 dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);

  for (i = 0; i < dcp->opt_2d.cnt; i++) {                   /* option list */
    option = dcp->opt_2d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_direct_pic_dbl(source_dpic, source_type, dc_name,
				      &option_test, 1);

    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_direct_pic_dbl(source_dpic, source_type, dc_name,
				      &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    dcp->opt_2d.n[i].nr = option_test;
  }
}



/* Stores documentation channel composed samples. If a sample list has all
   its components, it is stored to disk and its flags are cleared.

Revision: 06.01.2003 E.Sack

Input : - source FFT
        - source type
        - pointer to subject string
        - pointer to documentation channel
Output: - samples stored
        - documentation channel updated
*/
static void
dc_1d_store_fft(char *dc_name, fft_def src_fft, int source_type, 
		char *subject, dc_pic *dcp)
{
  int    i, option, option_test;
  SRC_IS src_available;

  assert(dcp);
  assert(subject);

  for (i = 0; i < dcp->opt_1d.cnt; i++) {                   /* option list */
    option = dcp->opt_1d.n[i].nr;

    if (!option) {
      continue;
    }

    option_test = option;
    src_available = dc_1d_ccf(src_fft, source_type, dc_name, subject,
			      &option_test, 1);

    if (src_available != SRC_VALID_COMPLETE) {
      assert(option_test == option);
      continue;                       /* preview: no success on completion */
    }

    option_test = option;
    src_available = dc_1d_ccf(src_fft, source_type, dc_name, subject,
			      &option_test, 0);
    assert(src_available==SRC_VALID_COMPLETE);
    dcp->opt_1d.n[i].nr = option_test;
  }
}



/* Resets documentation channel and erases temporary images not further
   needed.

Revision: 29.12.2002 E.Sack

Input : - pointer to documentation channel
Output: - documentation channel updated
        - number of deleted sources
*/
static int
dc_reset(dc_pic *dcp)
{
  int i, req_idx, src_type, del = 0, depends;

  assert(dcp);

  for (i = 0; i < dcp->dpl.cnt; i++) {
     src_type = dcp->dpl.dps[i].styp;

     if (!src_type) {                                      /* not a source */
       continue;
     }

     depends = dc_dependence(dcp->opt_2d, src_type);

     if (depends) {
       continue;                           /* source required for overlays */
     }

     req_idx = dc_exist_required(dcp->dpl, src_type);

     if (req_idx >= 0) {
       continue;                      /* source required for documentation */
     }

     dc_delete_source(&dcp->dpl, i);
     del++;
  }

  return del;
}



/* Stores documentation channel image source only for overlays. If the 
   image source is needed for an overlay, it is stored to disk or 
   internally.

Revision: 27.12.2002 E.Sack

Input : - image source
        - source type
        - pointer to documentation channel
Output: - documentation channel updated
*/
static void
dc_store_source(char *dc_name, pic_def src_pic, int src_type, dc_pic *dcp)
{
  int  found, src_count, depends;
  char *tname = 0;

  if (dcp->dpl.sar.width || dcp->dpl.sar.height) {    /* check consistency */

    assert(dcp->dpl.sar.width == src_pic.ar.width);
    assert(dcp->dpl.sar.height == src_pic.ar.height);
  }
  else {                        /* every source image shall have same size */
    dcp->dpl.sar = src_pic.ar;          /* store general source attributes */
  }

  depends = dc_dependence(dcp->opt_2d, src_type);

  if (!depends) {
    return;                                         /* source not required */
  }

  found = dc_exist_source(dcp->dpl, src_type);
  assert(found<0);                          /* source generation is unique */

  tname = dc_base_name(dc_name, DC_EXT_2D);

  switch (src_type) {                            /* flag for source images */
  case DC_BIT_ORG:
    add_string(DC_OPT_ORG_FILE, &tname);
    break;
  case DC_BIT_BMP:
    add_string(DC_OPT_BMP_FILE, &tname);
    break;
  case DC_BIT_CLEAN:
    add_string(DC_OPT_CLEAN_FILE, &tname);
    break;
  case DC_BIT_GV:
    add_string(DC_OPT_GV_FILE, &tname);
    break;
  default:
    fprintf(stderr, "\nERROR: '0x%X' is not a DC source flag\n\n", src_type);
    assert(0);                       /* service missing: not a source flag */
    break;
  }

  found = dc_exist_name(dcp->dpl, tname);                    /* check name */

  if (found >= 0) {                               /* source already stored */
    dcp->dpl.dps[found].styp = src_type;              /* source flag added */ 
    FREEMEM(tname);
    return;
  }

  src_count = dcp->dpl.cnt++;

  if (src_count) {
    REALLOC(dc_pic_src, dcp->dpl.cnt, dcp->dpl.dps);
  }
  else {
    CALLOC(dc_pic_src, dcp->dpl.cnt, dcp->dpl.dps);
  }

  init_dc_pic_src(&dcp->dpl.dps[src_count]);

  dcp->dpl.dps[src_count].styp = src_type;

  dc_store_type(dcp);                              /* calculate store type */

  if (dcp->dpl.store_sint & src_type) {             /* store source intern */
    CALLOC(pic_def, 1, dcp->dpl.dps[src_count].tpic); 
    init_pic_def(dcp->dpl.dps[src_count].tpic);
    copy_pic_def(&src_pic, dcp->dpl.dps[src_count].tpic);
  }
  else {                                            /* store source extern */
    dump_pic_def(src_pic, tname, src_pic.ar, 0, 0, xctrl.cfg.verbose>BE_QUIET,
		 PNM_FORMAT_UNDEF);
    add_string(tname, &dcp->dpl.dps[src_count].tname);
  }

  FREEMEM(tname);
}



/* Calculates store type of sources.

Revision: 11.01.2003 E.Sack

Input:  - option list
        - pointer to store type
Output: - documentation channel updated
Postcondition: - store types remains unchanged if it is set.
*/
static void
dc_store_type(dc_pic *dcp)
{
  int      i, bit_cnt, bit_no, bit_cnt_max, bit_no_max, size, bit;
  int      src_bits = DC_BIT_ORG | DC_BIT_BMP | DC_BIT_CLEAN | DC_BIT_GV;
  int2_lst *dod = 0;

  assert(dcp);

  if (!dcp->dpl.store_size) {
    return;                                          /* no internal stores */
  }
  else {
    size = dcp->dpl.store_size;
  }

  if (dcp->dpl.store_sint) {
    return;                                      /* store flag already set */
  }

  dod = &dcp->opt_2d;                                         /* short cut */

  while (size-- > 0) {
    bit_cnt_max = 0;
    bit_no_max = 0;

    for (bit = 0; bit < 32; bit++) {                      /* test bit list */
      bit_no = 1<<bit;

      if (!(src_bits & bit_no)) {
	continue;                                    /* only source images */
      }

      if (dcp->dpl.store_sint & bit_no) {           /* flag already stored */
	continue;
      }

      bit_cnt = 0;

      for (i = 0; i < dod->cnt; i++) {            /* investigate flag list */
	bit_cnt += dod->n[i].nr & bit_no;
      }

      if (bit_cnt_max < bit_cnt) {               /* save maximal bit count */
	bit_cnt_max = bit_cnt;
	bit_no_max = bit_no;
      }
    }

    dcp->dpl.store_sint |= bit_no_max;   /* store flag of most used source */
  }
}



/* Assembles base name for documentation images.

Revision: 29.12.2002 E.Sack

Input : - pointer to input filename string
Output: - DC base name
*/
static char*
dc_base_name(char *filename, char *type)
{
  char *dc_name = NULL;

  cat_string(filename, DC_NAME_EXT, type, &dc_name);

  return dc_name;
}



/* Looks for a name in a source list.

Revision: 29.12.2002 E.Sack

Input : - source list
        - name
Output: - source index (-1: no source found)
*/
static int
dc_exist_name(dc_pic_lst dpl, char *name)
{
  int i, d;

  assert(name);

  for (i = 0; i < dpl.cnt; i++) {
    if (!dpl.dps[i].tname) {
      continue;
    }

    d = strcmp(name, dpl.dps[i].tname);

    if (!d) {                                             /* equal strings */
      return i;
    }
  }

  return -1;
}



/* Load documentation channel with images generated by algorithm.

Revision: 30.12.2002 E.Sack

Input : - image source
        - source type
Output: - documentation channel updated
*/
void
dc_source_pic(char *dc_name, pic_def source_pic, int source_type)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  /* use new image source directly may finish some overlays                */
  dc_store_pic(dc_name, source_pic, source_type, pic_channel);

  /* delete stored sources no more needed                                  */
  dc_reset(pic_channel);

  /* store sources needed for unfinished overlays                          */
  dc_store_source(dc_name, source_pic, source_type, pic_channel);
}



/* Checks DC file name.

Revision: 21.04.2004 E.Sack

Input : - pointer to string
Output: - error message as flagged
        - flag (0: no error, 1: error)
*/
static int
check_dc_name(char *dcname, char *desc)
{
  int err;

  err = check_name(dcname);

  if (err) {
    string_to_log(main_log, "cannot source DC file name ", desc?desc:"", 
		  "\n");
  }

  return err;
}



/* Load documentation channel with RB, generated in code.

Revision: 30.12.2002 E.Sack

Input : - RB source
Output: - documentation channel updated
*/
void
dc_source_rb(char *dc_name, rb_lst source_rb)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  dc_store_rb(dc_name, source_rb, pic_channel);

  dc_reset(pic_channel);
}



/* Load documentation channel with MRB, generated in code.

Revision: 31.12.2002 E.Sack

Input : - MRB source
Output: - documentation channel updated
*/
void
dc_source_mrb(char *dc_name, rb_lst source_mrb)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  dc_store_mrb(dc_name, source_mrb, pic_channel);

  dc_reset(pic_channel);
}



/* Load documentation channel with candidate pattern, generated in code.

Revision: 31.12.2002 E.Sack

Input : - candidate list source
Output: - documentation channel updated
*/
void
dc_source_pattern_rb(char *dc_name, cand_lst source_cand)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  dc_store_pattern(dc_name, source_cand, pic_channel);

  dc_reset(pic_channel);
}



/* Load documentation channel with candidate pattern, generated in code.

Revision: 31.12.2002 E.Sack

Input : - candidate list source
        - pointer to classes
Output: - documentation channel updated
*/
void
dc_source_pattern_nb(char *dc_name, cand_lst source_cand, cls_def *source_cls)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  assert(source_cls);

  dc_store_nb(dc_name, source_cand, source_cls, pic_channel);

  dc_reset(pic_channel);
}



/* Load documentation channel with all online data sources generated in code
   except storable data.

Revision: 31.12.2002 E.Sack

Input : - RB source
        - MRB source
        - candidate list source
        - pointer to classes
        - horizontal line list source
Output: - documentation channel updated
*/
void
xdc_source_all(char *dc_name, rb_lst src_rb, rb_lst src_mrb, 
	       cand_lst src_cand, cls_def *src_cls, int_lst src_hline)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  assert(src_cls);

  dc_store_all(dc_name, src_rb, src_mrb, src_cand, src_cls, src_hline, 
	       pic_channel);

  dc_reset(pic_channel);
}



/* Load documentation channel with candidate pattern, generated in code.

Revision: 31.12.2002 E.Sack

Input : - horizontal line list source
Output: - documentation channel updated
*/
void
dc_source_hline(char *dc_name, int_lst source_hline)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  dc_store_hline(dc_name, source_hline, pic_channel);

  dc_reset(pic_channel);
}



/* Load documentation channel with samples, generated in code.

Revision: 06.01.2003 E.Sack

Input : - sample list source
        - source type
        - pointer to subject string
Output: - documentation channel updated
*/
void 
dc_source_double(char *dc_name, double_lst source_fctn, int source_type, 
		 char *subject)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  assert(subject);

  dc_1d_store_double(dc_name, source_fctn, source_type, subject, pic_channel);
}



/* Load documentation channel with images generated by algorithm.

Revision: 10.01.2003 E.Sack

Input : - pointer to input filename string
        - image
        - image type
Output: - documentation channel updated
*/
void
dc_source_pic_dbl(char *dc_name, pic_dbl source_dpic, int source_type)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  dc_store_pic_dbl(dc_name, source_dpic, source_type, pic_channel);

  dc_reset(pic_channel);
}



/* Load documentation channel with fft, generated in code.

Revision: 06.01.2003 E.Sack

Input : - source FFT
        - source type
        - pointer to subject string
Output: - documentation channel updated
*/
void 
dc_source_fft(char *dc_name, fft_def source_fft, int source_type, 
	      char *subject)
{
  int    err;
  dc_pic *pic_channel = xdc_get_channel();

  err = check_dc_name(dc_name, "");

  if (err) {
    return;
  }

  assert(subject);

  dc_1d_store_fft(dc_name, source_fft, source_type, subject, pic_channel);
}



/* Delivers an image with RB frames.

Revision: 05.01.2003 E.Sack

Input : - RB list
        - pointer to color
Output: - image
*/
static pic_def
dc_copy_frames(rb_lst rbl, col_def *color)
{
  int     y, x, y_start, x_start, y_end, x_end, number_rb; 
  int     left, right, upper, lower;
  pic_def pout;

  init_pic_def(&pout);
  pout.ar = rbl.rar;

  if (color) {
    CALLOC(col_def, AREA_SIZE(pout.ar), pout.col);
  }
  else {
    CALLOC(PICT, AREA_SIZE(pout.ar), pout.mem);
  }

  for (number_rb = 0; number_rb < rbl.cnt; number_rb++) {
    x_start = rbl.rb[number_rb].sc.column.start;
    x_end   = rbl.rb[number_rb].sc.column.end;
    y_start = rbl.rb[number_rb].sc.line.start;
    y_end   = rbl.rb[number_rb].sc.line.end;

    for (y = y_start+1; y < y_end; y++) {                    /* plot frame */
      left = x_start + y*rbl.rar.width;                 /* left frame side */
      right = x_end + y*rbl.rar.width;                 /* right frame side */

      if (color) {
	pout.col[left] = pout.col[right] = *color;
      }
      else {
	pout.mem[left] = pout.mem[right] = FOREGROUND;
      }
    }

    y_start *= rbl.rar.width;
    y_end   *= rbl.rar.width;

    for (x = x_start; x <= x_end; x++) {                     /* plot frame */
      upper = x + y_start;                             /* upper frame side */
      lower = x + y_end;                               /* lower frame side */

      if (color) {
	pout.col[upper] = pout.col[lower] = *color;
      }
      else {
	pout.mem[upper] = pout.mem[lower] = FOREGROUND;
      }
    }
  }

  return pout;
}



/* Extracts pattern regarding recognition results.

Revision: 31.12.2002 E.Sack

Input : - image
        - candidate list
        - pattern attribute
Output: - image containing attributed pattern
*/
static pic_def
dc_cinderella_pattern(pic_def pin, cand_lst hlist, PATTERN_ATT att,
		      PIXEL_KEY grey_backgnd)
{
  int        j, nbn, grey;
  pic_def    pout;
  sector_def clip;
  area_def   ar;
  col_def    fail_color = PATTERN_FAIL_COLOR, pattern_color;
  col_def    good_color = PATTERN_GOOD_COLOR;
  PNM_FORMAT pnm;

  assert(att == PATTERN_REC || att == PATTERN_UREC || att == PATTERN_ALL);

  pattern_color = (att==PATTERN_REC) ? good_color : fail_color;

  init_pic_def(&pout);
  pout.ar = pin.ar;

  pnm = automatic_format(pin, pin.ar, 0, 0, 0);
  grey = (pnm == P5) || (pnm == P2);

  if (pin.mem) {
    assert(!pin.col);
    CALLOC(PICT, AREA_SIZE(pout.ar), pout.mem);
    if (grey && grey_backgnd == FOREGROUND) {         /* invert background */
      set_bitmap(AREA_SIZE(pout.ar), pout.mem, PICT_MAX);
    }
  }
  else {
    assert(pin.col);
    assert(!pin.mem);
    CALLOC(col_def, AREA_SIZE(pout.ar), pout.col);
    }

  for (j = 0; j < hlist.cnt; j++) {               /* search candidate of RB */
    nbn = hlist.c[j].n;                                        /* NB number */

    if ((nbn < 0 && att == PATTERN_UREC) ||         /* unrecognized pattern */
	(nbn >= 0 && att == PATTERN_REC) ||           /* recognized pattern */
        (att == PATTERN_ALL)) {        /* unrecognized & recognized pattern */
      clip = hlist.c[j].knl;
      SECTOR_TO_AREA(clip, ar);

      if (pin.mem) {
	copy_bmps(pin, pout, ar, clip.column.start, clip.line.start, 
		  clip.column.start, clip.line.start);          /* 1:1 copy */
      }
      else {
	copy_color(pin, pout, ar, clip.column.start, clip.line.start, 
		   clip.column.start, clip.line.start);
      }
    }
  }

  return pout;
}



/* Assembles recogized NB.

Revision: 31.12.2002 E.Sack

Input : - document image size
        - candidate list
        - pointer to classes
Output: - image containing recogized NB
*/
static pic_def
dc_cinderella_nb(area_def doc_ar, cand_lst hlist, cls_def *cls)
{
  int      i, cnr, nnr;
  area_def ar;
  pic_def  pout;

  assert(cls);

  init_pic_def(&pout);
  pout.ar = doc_ar;
  CALLOC(PICT, AREA_SIZE(pout.ar), pout.mem);

  for (i = 0; i < hlist.cnt; i++) {                       /* recognized NB */
    if (hlist.c[i].n < 0) {                        /* unrecognized pattern */
      continue;
    }          

    SECTOR_TO_AREA(hlist.c[i].knl, ar);                     /* within image */
    cnr = hlist.c[i].c;
    nnr = hlist.c[i].n;

    copy_bmps(cls[cnr].nbl.nb[nnr].pat, pout, ar, 
	      hlist.c[i].knl.column.start-hlist.c[i].hdx.column.start,/* sx */
	      hlist.c[i].knl.line.start -hlist.c[i].hdx.line.start,   /* sy */
	      hlist.c[i].knl.column.start,                            /* dx */
	      hlist.c[i].knl.line.start);                             /* dy */
  }

  return pout;
}



/* Delivers a image with colored lines.

Revision: 02.01.2003 E.Sack

Input : - image size
        - pointer to color
        - line start point list
Output: - image
*/
static pic_def
dc_copy_lines(area_def ar_in, col_def *color, int_lst line)
{
  int     line_no, x, y, width = ar_in.width, height = ar_in.height;
  pic_def pout;

  init_pic_def(&pout);
  pout.ar = ar_in;

  if (color) {
    CALLOC(col_def, AREA_SIZE(pout.ar), pout.col);
  }
  else {
    CALLOC(PICT, AREA_SIZE(pout.ar), pout.mem);
  }

  for (line_no = 0; line_no < line.cnt; line_no++) {
    y = line.nr[line_no];
    assert(y < height);
    y *= width;

    for (x = 0; x < width; x++) {                  /* plot horizontal line */
      if (color) {
	pout.col[x+y] = *color;
      }
      else {
	pout.mem[x+y] = FOREGROUND;
      }
    }
  }

  return pout;
}



/* Delivers a image with colored RB subdivision lines.

Revision: 05.01.2003 E.Sack

Input : - RB list
        - pointer to color
Output: - image
*/
static pic_def
dc_copy_grid(rb_lst rbl, col_def *color)
{
  int     rnr, line_no, x, y, rb_start, rb_end;
  int     pic_width = rbl.rar.width, pic_height = rbl.rar.height;
  pic_def pout;

  init_pic_def(&pout);
  pout.ar = rbl.rar;

  if (color) {
    CALLOC(col_def, AREA_SIZE(pout.ar), pout.col);
  }
  else {
    CALLOC(PICT, AREA_SIZE(pout.ar), pout.mem);
  }

  for (rnr = 0; rnr < rbl.cnt; rnr++) {
    for (line_no = 1; line_no < rbl.rb[rnr].gridy.cnt; line_no++) {
      y = rbl.rb[rnr].gridy.ival[line_no].start;   /* only non-frame lines */
      assert(y < pic_height && y >= 0);
      y *= pic_width;

      rb_start = rbl.rb[rnr].sc.column.start;
      rb_end   = rbl.rb[rnr].sc.column.end;
      for (x = rb_start; x <= rb_end; x++) {    /* plot horizontal RB line */
	if (color) {
	  pout.col[x+y] = *color;
	}
	else {
	  pout.mem[x+y] = FOREGROUND;
	}
      }
    }
  }

  return pout;
}



dc_pic P_dchannel;                  /* (data hiding) documentation channel */



/* Initiate the documentation channel. Required for data hiding.

Revision: 01.11.2003 E.Sack

Input : - void
Output: - documentation channel initiated
Precondition: - to re-init at first call the appropriate dc_exit function.
*/
void
xplab_dc_init(void)
{
  init_dc_pic(&P_dchannel);
}



/* Empties documentation channel.

Revision: 01.01.2002 E.Sack

Input : - verbose flag
Output: - documentation channel updated
*/
void
xplab_dc_exit(void)
{
  int del1, del2, err = 0, verbose = xctrl.cfg.verbose > VERBOSE_NORMAL;

  del1 = dc_reset(&P_dchannel);

  if (del1 && verbose) {
    int_to_log(main_log, "WARNING-DC: ", del1, " option");
    string_to_log(main_log, del1==1?"":"s", " not reset earlier", "\n");
    err = 1;
  }

  del1 = dc_delete_flags(P_dchannel.opt_1d);
  del1 += dc_delete_flags(P_dchannel.opt_2d);
  del2 = dc_reset(&P_dchannel);
  assert(del1>=del2);

  if (del1 && verbose) {
    int_to_log(main_log, "WARNING-DC: ", del1, " option");
    string_to_log(main_log, del1==1?"":"s", " couldn't served", "\n");
    err = 1;
  }

  if (del2 && verbose) {
    int_to_log(main_log, "WARNING-DC: ", del2, " source");
    string_to_log(main_log, del2==1?"":"s", " couldn't served", "\n");
    err = 1;
  }

  if (verbose && !err) {
    string_to_log(main_log, "DC: service alright", "", "\n");
  }

  free_dc_pic(&P_dchannel);                      /* release all data in DC */
}



/* Access function for the documentation channel.

Revision: 08.04.2004 E.Sack

Input : - -
Output: - pointer to the documentation channel
*/
static dc_pic*
xdc_get_channel(void)
{
  xdc_assemble_sample(); /* real reason to hide channel - cannot be forget */

  return &P_dchannel;
}



/* Sets/Resets option flags.

Revision: 08.04.2004 E.Sack

Input : - -
Output: - flags updated
Precondition: - the flags can be set at any time, hence not synchronously to
                the current DC channel's data flow. The result may be 
                missed data (images) for output.
*/
static void
xdc_assemble_sample(void)
{
  update_int2_lst(xctrl.cfg.dc_1d, &P_dchannel.opt_1d, 1);
  update_int2_lst(xctrl.cfg.dc_2d, &P_dchannel.opt_2d, 1);
}

