/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_maintenance_be.h"
#include "xplab_training_be.h"
#include "xplab_image.h"
#include "xplab_tprint.h"
#include "xplab_batch.h"

/* local prototypes ********************************************************/
static void ref_nb(cls_lst clst, int cnr, int nb_from, int nb_to);
static int unref_nb_chn_lst(nb_chn_lst *ncl);
static int delete_unreferenced(cls_lst clst, int *deleted_nb);
static int unique_tps_id(tps_lst tpl, char **ident);

/* Functions ***************************************************************/

/* Checks the identities of two patterns.

Input : - NB
Output: - flag (0: pattern not identic, 1: pattern identic)
*/
int
compare_bmp(nb_def nb1, nb_def nb2)
{
  int l, len;

  if (nb1.pat.ar.width  != nb2.pat.ar.width || 
      nb1.pat.ar.height != nb2.pat.ar.height) {
    return 0;                    /* not identic because of different sizes */
  }

  len = nb1.pat.ar.width * nb1.pat.ar.height;

  for (l = 0; l < len; l++) {
    if (nb1.pat.mem[l] != nb2.pat.mem[l]) {
      return 0;                /* not identic because of different pattern */
    }
  }

  return 1;                                              /* no differences */
}



/* Calculates the width and height of the NB pattern of one class, 
   inclusive space left.
   The format is: one NB above the other in one column.

Revision: 21.09.2002 E.Sack

Input : - NB (pattern) list
        - vertical and horizontal space
        - pointer to area size
Output: - size updated as flagged
        - flag (0: size not updated, 1: size updated)
*/
int
id_col_nb(nb_lst nbl, int space_x, int space_y, area_def *ar_out)
{
   int i, max_w, max_l;

   assert(space_x >= 0); 
   assert(space_y >= 0);
   assert(ar_out);

   if (!nbl.nb) {                                    /* consistency check */
     assert(!nbl.cnt);
     return 0;
   }
   else {
     assert(nbl.cnt>0);
   }

   for (i = 0, max_w = 0, max_l = 0; i < nbl.cnt; i++) {
     max_w = XPMAX(nbl.nb[i].pat.ar.width, max_w);
     max_l += nbl.nb[i].pat.ar.height + 2*space_y;
   }

   ar_out->width = max_w + 2*space_x;    /* every cell with a frame around */
   ar_out->height = max_l;

   return 1;
}



/* Assembles an image of all NBs of a class. The format is: left to the 
   right, top to bottom. Virtually grid used has floating horizontal cell
   width and fixed cell height.

Revision: 25.08.2005 E.Sack

Input : - class
        - preferred form width
        - vertical and horizontal space around every pattern
Output: - image (may be empty)
        - NB auxiliary items set
Postcondition: - only printable NBs are used.
               - the auxiliary NB item points to the image output for a simple
                 handling of its references in that image. The caller must
                 be aware for freeing the output accordingly, what is made
                 using XPLABs freeing control mechanism.
               - the NB auxiliary items define some references against the
                 image output and hence are overwritten.
Usage : There is auxiliary data in every NB set to its location in the output
        image. This is done for convenience to keep track of the NB pattern,
        however not necessarily used. The caller must be aware that the
        auxiliary NB items are overwritten.
*/
pic_def
xplab_sample_image(cls_def cls, int form_width, int space_x, int space_y)
{
  int     f, z, i, maxw, maxh, updated, width, height;
  pic_def pic;

  init_pic_def(&pic);

  if (!cls.att.printable) {
    return pic;                                       /* class not printable */
  }

  updated = id_col_nb(cls.nbl, space_x, space_y,          /* size estimation */
		      &pic.ar);

  if (!updated) {
    return pic;                                               /* class empty */
  }

  maxw = (int)sqrt((double)AREA_SIZE(pic.ar)) + 1;     /* square form factor */
  assert(maxw);

  maxw = XPMAX(maxw, form_width);             /* override square form factor */

  /* coarse image width                                                      */
  for (i = 0, pic.ar.width = 0; i < cls.nbl.cnt; i++) {
    width = cls.nbl.nb[i].pat.ar.width + 2*space_x;

    if (pic.ar.width + width <= maxw) {
      pic.ar.width += width;
    }
    else {
      break;
    }
  }

  for (i = 0, maxh = 0, maxw = 0; i < cls.nbl.cnt; i++) {   /* finely tuning */
    maxw = XPMAX(maxw, cls.nbl.nb[i].pat.ar.width + 2*space_x);/* biggest NB */
    maxh = XPMAX(maxh, cls.nbl.nb[i].pat.ar.height + 2*space_y);
  }

  pic.ar.width = XPMAX(pic.ar.width, maxw);             /* final image width */

  /* fixing of the image height                                              */
  for (i = maxw = pic.ar.height = 0; i < cls.nbl.cnt; i++) {
    width = cls.nbl.nb[i].pat.ar.width + 2*space_x;

    if (maxw + width > pic.ar.width) {
      pic.ar.height += maxh;
      maxw = 0;
    }

    maxw += width;
  }

  pic.ar.height += maxh;                 /* don't forget last row of pattern */
  CALLOC(PICT, AREA_SIZE(pic.ar), pic.mem);                         /* image */

  for (i = maxw = 0, z = space_y; i < cls.nbl.cnt; i++) { /* enroll patterns */
    width  = cls.nbl.nb[i].pat.ar.width  + 2*space_x;
    height = cls.nbl.nb[i].pat.ar.height + 2*space_y;

    if (maxw + width > pic.ar.width) {
      z += maxh;
      maxw = 0;
    }

    switch (cls.att.alignment) {
    case UPPERLINE : f = 0;
                     break;
    case MIDDLELINE: f = (maxh - height) >> 1;
                     break;
    default:         f = maxh  - height;
                     break;
    }

    copy_bmps(cls.nbl.nb[i].pat, pic, cls.nbl.nb[i].pat.ar, 0, 0,
	      maxw + space_x, z + f);

    /* auxiliary: calculate offset of absolute location of NB in output image*/
    cls.nbl.nb[i].dat.afs.x = maxw + space_x;                 /* save offset */
    cls.nbl.nb[i].dat.afs.y = z + f;

    /* auxiliary: calculate absolute location of NB in the output image      */
    AREA_TO_SECTOR(cls.nbl.nb[i].pat.ar, cls.nbl.nb[i].dat.sec);

    cls.nbl.nb[i].dat.sec.column.start += cls.nbl.nb[i].dat.afs.x - space_x;
    cls.nbl.nb[i].dat.sec.column.end   += cls.nbl.nb[i].dat.afs.x + space_x;
    cls.nbl.nb[i].dat.sec.line.start   += cls.nbl.nb[i].dat.afs.y - space_y;
    cls.nbl.nb[i].dat.sec.line.end     += cls.nbl.nb[i].dat.afs.y + space_y;

    ASSERT_SECTOR(cls.nbl.nb[i].dat.sec, pic.ar);    /* test NB within image */

    maxw += width;
  }

  return pic;
}



/* Colors patterns in a sector.

Revision: 28.09.2004 E.Sack

Input : - pointer to color image memory
        - labelled image
        - color definition
        - focus image sector
        - flag (0: change sector outside, else: don't change sector outside)
Output: - pattern updated
Precondition:  - color image memory may reside in the labelled image structure.
Postcondition: - clipping overwritten
*/
void /*TBD diese Fktn. gehört nach xplab_training_be.c */
xprint_rb(col_def *pixbuf, pic_def mark_img, pattern_color col, sector_def sec,
	  int outside_unchanged) 
{
  int     yy, y, x;
  col_def *cpixel;

  assert(pixbuf);
  assert(mark_img.mem);

  for (y = 0; y < mark_img.ar.height; y++) {
    yy = y * mark_img.ar.width;

    for (x = 0; x < mark_img.ar.width; x++) {
      cpixel = &pixbuf[x + yy];                     /* image rendering marks */

      switch (mark_img.mem[x + yy]) {        /* query image containing marks */
      case BACKGROUND:/* does'nt really need marked pixels, just for rendering */
	*cpixel = WITHIN_SECTOR(x,y,sec) ? col.bgd_in : 
	                     (outside_unchanged ? *cpixel : col.bgd_out);
	break;

      case FOREGROUND:           *cpixel = col.fgd; break;
      case XPLAB_SELECTED_PIXEL: *cpixel = col.sel; break;
      case XPLAB_SELECTBL_PIXEL: *cpixel = col.sbl; break;
      case XPLAB_TRAINED_PIXEL:  *cpixel = col.tnd; break;
      case XPLAB_RESERVED_PIXEL: *cpixel = col.res; break;
      case XPLAB_IGNORED_PIXEL:  *cpixel = col.ign; break;
      default: assert(0); break;
      }
    }
  }
}



/* Converts coordinates to NB index.

Revision: 26.08.2005 E.Sack

Input : - class
        - coordinates x,y
        - spacing x,y
Output: - NB index (-1: no NB found)
Usage:  - Counterpart of function xid_nb_image().
*/
int
coordinates_to_index(cls_def cls, int x, int y)
{
  int nnr;

  if (!cls.nbl.nb) {                                    /* consistency check */
    assert(!cls.nbl.cnt);
    return -1;                                                /* class empty */
  }
  else {
    assert(cls.nbl.cnt>0);
  }

  /* the coordinates are converted into a NB index                           */
  for (nnr = 0; nnr < cls.nbl.cnt; nnr++) {
    if (SECTOR_SIZE(cls.nbl.nb[nnr].dat.sec) <= 0) {
      continue;                                         /* no sector defined */
    }

    if (WITHIN_SECTOR(x, y, cls.nbl.nb[nnr].dat.sec)) {    /* inside NB area */
      return nnr;                                                /* NB index */
    }
  }

  return -1;       /* no NB index found: inconsistence of class and NB image */
}



/* Deletes superfluous NB.

Revision: 14.09.2002 E.Sack

Input : - class list
        - pointer to sum of deleted NB
Output: - class list updated
        - sum of deleted chains
        - sum of deleted NB updated
Precondition: - pointer to sum of deleted NB may be NULL
*/
static int
delete_unreferenced(cls_lst clst, int *deleted_nb) 
{
  int     cls_nr, nb_nr, changed = 1, deleted_refs = 0;
  nb_def  nb;
  cls_def class;

  while (changed) {
    if (!clst.cls) {                                  /* consistency check */
      assert(!clst.cnt);
      return deleted_refs;                                      /* no data */
    }
    else {
      assert(clst.cnt>0);
    }

    for (cls_nr = changed = 0; cls_nr < clst.cnt; cls_nr++) {
      class = clst.cls[cls_nr];                       /* shortcut to class */

      if (!class.nbl.nb) {                            /* consistency check */
	assert(!class.nbl.cnt);
      }
      else {
	assert(class.nbl.cnt>0);
      }

      if (class.att.printable) {
	continue;                                          /* printable NB */
      }

      for (nb_nr = 0; nb_nr < class.nbl.cnt; nb_nr++) {
	nb = class.nbl.nb[nb_nr];                        /* shortcut to NB */

	if (!nb.cnb.cd) {                             /* consistency check */
	  assert(!nb.cnb.cnt);
	}
	else {
	  assert(nb.cnb.cnt>0);
	}

	if (nb.cnb.cnt) {
	  continue;                                          /* chained NB */
	}

	deleted_refs = erase_nb_def(clst, cls_nr, nb_nr, deleted_nb, 1);
	changed = 1;
	break;
      }

      if (changed) {
	break;                                               /* DB changed */
      }
    }
  }

  return deleted_refs;
}



/* Deletes negative indexes in a class list contained in a list.

Revision: 14.09.2002 E.Sack

Input : - document class list
        - pointer to sum of deleted NB
Output: - class list updated
        - sum of deleted chains
        - sum of deleted NB updated
Precondition: - pointer to sum of deleted NB may be NULL
Postcondition: - document class list may be empty
*/
int
unref_indexes(cls_lst clst, int *deleted_nb) 
{
  int     cls_nr, nb_nr, changed = 1, sum = 0;
  nb_def  *nb = 0;
  cls_def class;
   
  while (changed) {
    if (!clst.cls) {                                  /* consistency check */
      assert(!clst.cnt);
      return sum;                                               /* no data */
    }
    else {
      assert(clst.cnt>0);
    }

    for (cls_nr = changed = 0; cls_nr < clst.cnt; cls_nr++) {
      class = clst.cls[cls_nr];                       /* shortcut to class */

      if (!class.nbl.nb) {                            /* consistency check */
	assert(!class.nbl.cnt);
      }
      else {
	assert(class.nbl.cnt>0);
      }

      for (nb_nr = 0; nb_nr < class.nbl.cnt; nb_nr++) {         
	nb = &class.nbl.nb[nb_nr];                       /* shortcut to NB */

	changed = unref_nb_chn_lst(&nb->cnb);
	changed += unref_nb_chn_lst(&nb->nnb);
	sum += changed;

	if (changed) {
	  break;                                             /* DB changed */
	}       
      }

      if (changed) {
	break;                                               /* DB changed */
      }       
    }
  }

  sum += delete_unreferenced(clst, deleted_nb);

  return sum;
}



/* Deletes negative indexes in a NB chain list.

Revision: 14.09.2002 E.Sack

Input : - pointer to NB chain list
Output: - NB chain list updated
        - flag (0: no changes, 1: changes)
Postcondition: - NB chain list may be empty
*/
static int
unref_nb_chn_lst(nb_chn_lst *ncl)
{
  int        ncd_nr;
  nb_chn_def chain;

  assert(ncl);

  if (!ncl->cd) {                                     /* consistency check */
    assert(!ncl->cnt);
    return 0;                                                   /* no data */
  }
  else {
    assert(ncl->cnt>0);
  }

  for (ncd_nr = 0; ncd_nr < ncl->cnt; ncd_nr++) {
    chain = ncl->cd[ncd_nr];                          /* shortcut to chain */

    if (chain.cdx < 0 || chain.ndx < 0) {
      delete_nb_chn_def(ncl, ncd_nr);
      return 1;
    }
  }

  return 0;
}



/* Changes a NB reference.

Revision: 14.09.2002 E.Sack

Input : - document class list
        - class number
        - old NB number
        - new NB number
Output: - document class list updated
*/
static void
ref_nb(cls_lst clst, int cnr, int nb_from, int nb_to) 
{
  int     cls_nr, nb_nr;
  nb_def  nb;
  cls_def class;

  if (nb_from == nb_to) {                             /* nothing to change */
    return;
  }

  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
    return;                                                     /* no data */
  }
  else {
    assert(clst.cnt>0);
  }

  for (cls_nr = 0; cls_nr < clst.cnt; cls_nr++) {
    class = clst.cls[cls_nr];                         /* shortcut to class */

    if (!class.nbl.nb) {                              /* consistency check */
      assert(!class.nbl.cnt);
    }
    else {
      assert(class.nbl.cnt>0);
    }

    for (nb_nr = 0; nb_nr < class.nbl.cnt; nb_nr++) {         
      nb = class.nbl.nb[nb_nr];                          /* shortcut to NB */

      nref_nb_chn_lst(nb.cnb, cnr, nb_from, nb_to);             /* parents */
      nref_nb_chn_lst(nb.nnb, cnr, nb_from, nb_to);            /* children */
    }
  }
}



/* Changes a class/NB pair reference.

Revision: 08.06.2004 E.Sack

Input : - class list
        - old class number
        - new class number
        - old NB number
        - new NB number
Output: - class list updated
*/
void
ref_pair(cls_lst clst, int cls_from, int cls_to, int nb_from, int nb_to)
{
  int     cls_nr, nb_nr;
  nb_def  nb;
  cls_def class;

  if (cls_from == cls_to) {                           /* nothing to change */
    return;
  }

  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
  }
  else {
    assert(clst.cnt>0);
  }

  for (cls_nr = 0; cls_nr < clst.cnt; cls_nr++) {
    class = clst.cls[cls_nr];                         /* shortcut to class */

    if (!class.nbl.nb) {                              /* consistency check */
      assert(!class.nbl.cnt);
    }
    else {
      assert(class.nbl.cnt>0);
    }

    for (nb_nr = 0; nb_nr < class.nbl.cnt; nb_nr++) {         
      nb = class.nbl.nb[nb_nr];                          /* shortcut to NB */

      pref_nb_chn_lst(nb.cnb, cls_from, cls_to, nb_from, nb_to);/* parents */
      pref_nb_chn_lst(nb.nnb, cls_from, cls_to, nb_from, nb_to);/* children*/
    }
  }
}



/* Changes NB chain references.

Revision: 14.09.2002 E.Sack

Input : - NB chain list
        - class number
        - old NB number
        - new NB number
Output: - NB chain list updated
        - sum of changes
*/
int
nref_nb_chn_lst(nb_chn_lst ncl, int cnr, int nb_from, int nb_to)
{
  int        chn_nr, sum = 0;
  nb_chn_def *chain = 0;

  if (!ncl.cd) {                                      /* consistency check */
    assert(!ncl.cnt);
    return sum;                                                 /* no data */
  }
  else {
    assert(ncl.cnt>0);
  }

  for (chn_nr = 0; chn_nr < ncl.cnt; chn_nr++) {
    chain = &ncl.cd[chn_nr];                          /* shortcut to chain */

    if (chain->cdx == cnr && chain->ndx == nb_from) {
      chain->ndx = nb_to;
      sum++;
    }
  }

  return sum;
}



/* Changes class/NB pair chain references.

Revision: 24.09.2002 E.Sack

Input : - NB chain list
        - old class number
        - new class number
        - sum of changes
Output: - NB chain list updated
*/
int
pref_nb_chn_lst(nb_chn_lst ncl, int cls_from, int cls_to, int nb_from, 
		int nb_to)
{
  int        chn_nr, changed = 0;
  nb_chn_def *chain = 0;

  if (!ncl.cd) {                                      /* consistency check */
    assert(!ncl.cnt);
    return changed;                                             /* no data */
  }
  else {
    assert(ncl.cnt>0);
  }

  for (chn_nr = 0; chn_nr < ncl.cnt; chn_nr++) {
    chain = &ncl.cd[chn_nr];                          /* shortcut to chain */

    if (chain->cdx == cls_from && chain->ndx == nb_from) {
      chain->cdx = cls_to;
      chain->ndx = nb_to;
      changed++;
    }
  }

  return changed;
}



/* Deletes a NB chain.

Revision: 14.09.2002 E.Sack

Input : - pointer to NB chain entry list
        - NB chain entry number
Output: - NB chain entry list updated
Postcondition: - NB chain entry list may be empty
*/
void
delete_nb_chn_def(nb_chn_lst *ncl, int ncd_nr) 
{
  int last;

  assert(ncl);

  if (!ncl->cnt) {                                              /* no data */
    assert(!ncl->cd);  
    return;
  }
  else {
    assert(ncl->cnt>0);  
    assert(ncl->cd);  
  }

  last = ncl->cnt-1;
  assert(ncd_nr >= 0);
  assert(ncd_nr <= last);

  if (ncd_nr < last) {                               /* chain order change */
    copy_nb_chn_def(&ncl->cd[last], &ncl->cd[ncd_nr]);
  }

  ncl->cnt--;                                            /* one chain less */

  if (ncl->cnt) {                                              /* clean up */
    REALLOC(nb_chn_def, ncl->cnt, ncl->cd);
  }
  else {
    free_nb_chn_lst(ncl);
  }
}



/* Deletes a NB.

Revision: 22.12.2005 E.Sack

Input : - class list
        - class number
        - NB number
        - pointer to sum of deleted NB
        - verbose flag
Output: - class list updated
        - sum of deleted chains
        - sum of deleted NB updated
Precondition:  - pointer to sum of deleted NB may be NULL
Postcondition: - NB list may be empty
*/
int
erase_nb_def(cls_lst clst, int cnr, int nnr, int *deleted_nb, int vflag)
{
  int last, deleted_refs = 0;

  if (!clst.cnt) {                                              /* no data */
    assert(!clst.cls);
    return deleted_refs;
  }
  else {
    assert(clst.cnt>0);
    assert(clst.cls);
  }

  assert(cnr >= 0);
  assert(cnr < clst.cnt);

  assert(clst.cls[cnr].nbl.nb);
  assert(clst.cls[cnr].nbl.cnt>0);

  last = clst.cls[cnr].nbl.cnt-1;
  assert(last >= 0);
  assert(nnr <= last);

  if (vflag) {
    int_to_log(main_log, "erasing NB[", nnr, "]");
    int_to_log(main_log, " in class[", cnr, "]");
    string_to_log(main_log, " '", clst.cls[cnr].att.class, "'\n");
  }

  ref_nb(clst, cnr, nnr, -1);
  ref_nb(clst, cnr, last, nnr);                            /* index change */

  if (nnr < last) {                                     /* NB index change */
    copy_nb_def(&clst.cls[cnr].nbl.nb[last], &clst.cls[cnr].nbl.nb[nnr], FALSE);
  }

  free_nb_def(&clst.cls[cnr].nbl.nb[last]);

  clst.cls[cnr].nbl.cnt--;                                  /* one NB less */

  if (clst.cls[cnr].nbl.cnt) {
    REALLOC(nb_def, clst.cls[cnr].nbl.cnt, clst.cls[cnr].nbl.nb);
  }
  else {
    free_nb_lst(&clst.cls[cnr].nbl);
  }

  if (deleted_nb) {
    (*deleted_nb)++;
  }

  deleted_refs = unref_indexes(clst, deleted_nb);

  return deleted_refs;
}



/* Seeks class of a certain name.

Revision: 27.07.2002 E.Sack

Input : - pointer to class name
        - class list
Output: - class number (-1: no class found)
*/
int
search_class(const char *class_name, cls_lst clst)
{
  int c, err;

  err = check_name(class_name);
  assert(!err);

  for (c = 0; c < clst.cnt; c++) {
    if (!strcmp(clst.cls[c].att.class, class_name)) {
      return c;
    }
  }

  return -1;                                             /* no class found */
}



/* Searching the DB for a certain TPS ID.

Revision: 25.08.2002 E.Sack

Input : - pointer to TPS ID
        - TPS list
Output: - TPS index for TPS ID (-1: no ID found)
        - TPS ID updated
Precondition: - empty input TPS ID replaced with first ID found in DB if 
                there is any.
*/
int 
search_tps_id(char **ident, tps_lst tpl)
{
  int err, i, ident_available = 0, tps_id_available = 0;

  assert(ident);

  err = check_name(*ident);

  if (!err) {                                /* avoid using zero pointer */
    ident_available = 1;
  }

  if (tpl.tps) {
    assert(tpl.cnt>0);
  }
  else {
    assert(tpl.cnt==0);
    return -1;                                             /* no documents */
  }

  err = check_name(tpl.tps[0].ident);

  if (!err) {                                /* avoid using zero pointer */
    tps_id_available = 1;
  }
  else {
    assert(0);
  }

  if (!ident_available) {            /* use default TPS ID == first TPS ID */
    if (tps_id_available) {
	add_string(tpl.tps[0].ident, ident);
	return 0;
    }

    return -1;
  }

  for (i = 0; i < tpl.cnt; i++) {
    if (strcmp(tpl.tps[i].ident, *ident)) {
      continue;                                       /* other document ID */
    }

    return i;
  }

  return -1;                                            /* no TPS ID found */
}



/* Merges two class lists. As a class list has unique classes, classes are
   extended or newly created, but never doubled. Chain indexes are kept
   consistently.

Revision: 08.06.2004 E.Sack

Input : - input class list
        - pointer to output class list
Output: - output class list updated
*/
void
cat_cls_lst(cls_lst src, cls_lst *dst)
{
  int     cnr, nnr, nb_count, new_cnr, new_nnr_offset;
  cls_lst ctmp;

  assert(dst);

  init_cls_lst(&ctmp);

  if (!src.cls) {                                     /* check consistency */
    assert(!src.cnt);
    return;                                                     /* no data */
  }
  else {
    assert(src.cnt>0);
  }

  if (!dst->cls) {                                    /* check consistency */
    assert(!dst->cnt);
  }
  else {
    assert(dst->cnt>0);
  }

  copy_cls_lst(&src, &ctmp, FALSE);                /* protect source input */

  for (cnr = 0; cnr < ctmp.cnt; cnr++) {        /* chain referencing loop  */
    nb_count = ctmp.cls[cnr].nbl.cnt;

    if (!ctmp.cls[cnr].nbl.nb) {                      /* check consistency */
      assert(!nb_count);
      continue;                                                   /* no NB */
    }
    else {
      assert(nb_count>0);
    }

    for (nnr = 0; nnr < nb_count; nnr++) {      /* unique negative indexes */
      ref_pair(ctmp, cnr, cnr-ctmp.cnt, nnr, nnr-nb_count); /* referencing */
    }
  }

  for (cnr = 0; cnr < ctmp.cnt; cnr++) {                      /* copy loop */
    nb_count = ctmp.cls[cnr].nbl.cnt;
    new_cnr = cnr;                               /* default changing index */
    new_nnr_offset = 0;

    /* There is only one changing index so we have to reference the source.*/
    /* Hence the consequence is the source chains will be disturbed.       */
    cat_cls_def(ctmp.cls[cnr], dst, -1, &new_cnr, &new_nnr_offset);

    if (!nb_count) {
      continue;                                                   /* no NB */
    }

    /* source chains must have destination referen. because of future copy */
    for (nnr = 0; nnr < nb_count; nnr++) {     /* chain dereferencing loop */
      ref_pair(ctmp, cnr-ctmp.cnt, new_cnr,        /* source dereferencing */
	       nnr-nb_count, nnr+new_nnr_offset);

      ref_pair(*dst, cnr-ctmp.cnt, new_cnr,   /* destination dereferencing */
	       nnr-nb_count, nnr+new_nnr_offset);
    }
  }

  free_cls_lst(&ctmp);
}



/* Merges a class to a class list. As a class list has unique classes,
   classes are extended or newly created, but never doubled.

Revision: 08.06.2004 E.Sack

Input : - source class list
        - pointer to destination class list
        - destination class number to inhibit (negative to ignore)
        - pointer to new class number
        - pointer to new NB index
Output: - destination class list updated 
        - new class number updated
        - new NB index updated
Postcondition: - if destination has no existing class, the whole source
                 class is copied. Else only the NB list of the source class
                 is added to the destination NB list.
*/
void
cat_cls_def(cls_def src, cls_lst *dst, int inhibit_dst, int *new_cnr, 
	    int *new_nnr)
{
  int n;

  assert(dst);
  assert(new_nnr);
  assert(new_cnr);

  if (!dst->cls) {                                    /* check consistency */
    assert(!dst->cnt);
  }
  else {
    assert(dst->cnt>0);
  }

  *new_cnr = search_class(src.att.class, *dst);   /* classnames are unique */

  if (*new_cnr >= 0 && *new_cnr == inhibit_dst) {   /* avoid certain class */
    *new_cnr = -1;
  }

  if (*new_cnr >= 0) {                            /* extend existing class */
    if (!src.nbl.cnt) {
      return;                                            /* nothing to add */
    }

    *new_nnr = dst->cls[*new_cnr].nbl.cnt;         /* offset of NB indexes */
    dst->cls[*new_cnr].nbl.cnt += src.nbl.cnt;

    REALLOC(nb_def, dst->cls[*new_cnr].nbl.cnt, dst->cls[*new_cnr].nbl.nb);

    for (n = 0; n < src.nbl.cnt; n++) { 
      init_nb_def(&dst->cls[*new_cnr].nbl.nb[*new_nnr+n]);

      copy_nb_def(&src.nbl.nb[n], &dst->cls[*new_cnr].nbl.nb[*new_nnr+n], FALSE);
    }
  }
  else {	                                       /* create new class */
    *new_cnr = dst->cnt++;                             /* new class number */

    if (*new_cnr) {
      REALLOC(cls_def, dst->cnt, dst->cls);
    }
    else {
      CALLOC(cls_def, dst->cnt, dst->cls);
    }

    init_cls_def(&dst->cls[*new_cnr]);

    copy_cls_def(&src, &dst->cls[*new_cnr], FALSE);
    *new_nnr = 0;                           /* NNB chain indexes unchanged */
  }
}



/* Gives back a default setting.

Revision: 02.10.2003 E.Sack

Input : - -
Output: - default setting
*/
void
xdefault_media_att(media_att *med)
{
  assert(med);

  med->dpi = DPI_DEF;
  med->in_line.method = OVERLAPPING;
}



/* Adds a default empty TPS definition to a TPS list.

Revision: 14.02.2004 E.Sack

Input : - pointer to TPS list
        - pointer to TPS ID
Output: - TPS list updated
        - TPS ID updated
        - index of TPS (-1: no TPS added)
Precondition: - TPS ID may be NULL
Postcondition: - the TPS ID is made unique.
*/
int 
create_tps_def(tps_lst *tpl, char **ident)
{
  int err, found;

  assert(ident);

  if (!tpl->tps) {                                     /* consistency check */
    assert(!tpl->cnt);
  }
  else {
    assert(tpl->cnt>0);
  }

  if (tpl->cnt >= XPLAB_MAX_TPS) {
    int_to_log(main_log, "maximal amount (", XPLAB_MAX_TPS, 
	       ") of Trained Pattern Sets reached");
    return -1;
  }

  err = unique_tps_id(*tpl, ident);                  /* make TPS ID unique */
  if (err) {
    return -1;
  }

  found = tpl->cnt++;                                 /* begin allocation  */

  if (found) {                                              /* list exists */
    REALLOC(tps_def, tpl->cnt, tpl->tps);
  }
  else {                                             /* list doesn't exist */
    CALLOC(tps_def, tpl->cnt, tpl->tps);
  }

  init_tps_def(&tpl->tps[found]);

  add_string(*ident, &tpl->tps[found].ident);               /* copy TPS ID */

  return found;
}



/* Creates an unique, non empty TPS ID.

Revision: 14.02.2004 E.Sack

Input : - document list
        - pointer to TPS ID string
Output: - string updated
        - flag (0: string updated, 1: string not updated)
*/
static int
unique_tps_id(tps_lst tpl, char **ident)
{
  int  d, len, err = 0, uni = 0;
  char *base = NULL, *ustr = NULL;

  assert(ident);

  add_string(*ident, &ustr);                                 /* copy input */

  d = check_name(ustr);

  if (d) {
    add_string(TPSID_DEFAULT, &ustr);                           /* default */
  }

  add_string(ustr, &base);                               /* initial string */

  len = strlen(base);

  while (1) {
    if (len >= MAX_LEN_TPS_ID) {                  /* limited string length */
      int_to_log(main_log, "WARNING: TPS ID length (", len+1, ") ");/*incl. LF */
      int_to_log(main_log, "exceeds maximum (", MAX_LEN_TPS_ID, ")\n");
      err = 1;
      break;
    }

    d = search_tps_id(&ustr, tpl);       /* search string among the others */

    if (d < 0) {
      FREEMEM(*ident);
      add_string(ustr, ident);
      break;                                              /* unique string */
    }

    FREEMEM(ustr);                                         /* reset string */
    int_to_string(base, uni++, "", &ustr);                /* create string */
    len = strlen(ustr);
  }

  FREEMEM(ustr);
  FREEMEM(base);

  return err;
}



/* Checks auxiliary references.

Revision: 01.01.2006 E.Sack

Input : - TPS list
      . - flag (0: leave DB unchanged, 1: repair/clean DB)
Output: - error sum
        - DB updated as flagged
*/
int 
check_aux(tps_lst tpl, int clear)
{
  int        t, c, err = 0;
  char       *check_msg = NULL;
  DATA_STATE dstat;

  for (t = 0; t < tpl.cnt; t++) {
    for (c = 0; c < tpl.tps[t].cl.cnt; c++) {
      dstat = state_inh_lst(tpl.tps[t].cl.cls[c].inl);

      if (dstat != STATE_IS_INIT) {
	if (clear) {
	  free_inh_lst(&tpl.tps[t].cl.cls[c].inl);
	}
	else {
	  FREEMEM(check_msg);
	  check_message_class(err?"INFO: ":"\nINFO: ", t, c, &check_msg);
	  string_to_log(main_log, check_msg, 
			" - inhibition list defined\n", "");
	}

	err++;
      }
    }
  }

  FREEMEM(check_msg);

  return err;
}



/* Builds check message enumeration string.

Revision: 10.09.2004 E.Sack

Input : - pointer to prefix string
        - TPS index
        - class index
        - NB index
        - pointer to pointer to string
Output: - string updated
*/
void
check_message_nb(char *prefix, int tps_idx, int cls_idx, int nb_idx, 
		 char **string_out)
{
  assert(string_out);

  add_string(prefix, string_out);
  int_to_string("NB[", tps_idx, "]", string_out);
  int_to_string("[", cls_idx, "]", string_out);
  int_to_string("[", nb_idx, "]", string_out);
}



/* Builds check message enumeration string.

Revision: 10.09.2004 E.Sack

Input : - pointer to prefix string
        - TPS index
        - class index
        - pointer to pointer to string
Output: - string updated
*/
void
check_message_class(char *prefix, int tps_idx, int cls_idx, char **string_out)
{
  assert(string_out);

  add_string(prefix, string_out);
  int_to_string("Class[", tps_idx, "]", string_out);
  int_to_string("[", cls_idx, "]", string_out);
}



/* Builds check message enumeration string.

Revision: 11.09.2004 E.Sack

Input : - pointer to prefix string
        - TPS index
        - pointer to pointer to string
Output: - string updated
*/
void
check_message_tps(char *prefix, int tps_idx, char **string_out)
{
  assert(string_out);

  add_string(prefix, string_out);
  int_to_string("TPS[", tps_idx, "]", string_out);
}



/* Returns UTF8 converted string.

Revision: 30.07.2005 E.Sack

Input : - pointer to input string
Output: - pointer to UTF8 string
*/
gchar *
xplab_to_utf(const gchar *itxt)
{
  gchar  *sum;
  gsize  bytes_read;
  gsize  bytes_written;
  GError *error = NULL;

  sum = g_convert(itxt, -1, XPLAB_TO_CODESET, XPLAB_FROM_CODESET, &bytes_read,
		  &bytes_written, &error);

  if (error) {
    g_error_free(error);
    error = NULL;

    g_free(sum);
    sum = g_locale_to_utf8(itxt, -1, &bytes_read, &bytes_written, &error);

    if (error) {
      g_error_free(error);
      error = NULL;
      g_free(sum);
      sum = NULL;
    }
  }

  return sum;
}
