/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_selection.h"
#include "xplab_preferences_fe.h"
#include "xplab_image.h"
#include "xplab_dbio.h"
#include "xplab_replay.h"
#include "xplab_maintenance_fe.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/
#define XPLAB_SELECTION_TITLE_ITEM "load file"
#define XPLAB_SELECTION_REPLAY_COLOR {0, 0xFFFF,0xBEBE,0xBEBE}/* pink background */
#define XPLAB_SELECTION_IMAGE_COLOR {0, 0xF0F0,0xF0F0,0xD0D0}/* light yellow background */
#define XPLAB_SELECTION_DB_COLOR {0, 0xD6D6,0xE8E8,0xFFFF}/* light blue background */
#define XPLAB_SELECTION_MAX_HISTORY_FILES 5

typedef struct {
  GtkWidget *selection_dialog, *slabel, **caller;
  gboolean  browse;
} sctn_def;

static sctn_def se;                                      /* selection dialog */
static text_ring db_names;                       /* history for loaded files */
static text_ring image_names;                    /* history for loaded files */

/* local prototypes ********************************************************/
static void init_sctn_def(sctn_def *sd);
static void free_sctn_def(sctn_def *sd);

/* Functions assembling widgets */
static void xplab_selection(void);
static char *selection_load_file(char *filespec);
static void selection_buttons(GtkFileChooser *selection_dialog);
static void selection_filter(GtkFileChooser *selection_dialog);
static gboolean filter_db(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN);
static gboolean filter_image(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN);
static gboolean filter_session(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN);
static gboolean filter_replay(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN);

/* Handlers/Callbacks */
static void cb_selection_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID);

/* Functions assembling handlers/callbacks */
static int selection_scan_image(FILE *file);
static int selection_scan_session(FILE *file);
static int selection_scan_db(const char *db_name, FILE *file);
static int selection_scan_replay(const char *replay_name, FILE *file);
static void put_file_name(const char *source, text_ring *dest);

/* Functions ***************************************************************/

static void
init_sctn_def(sctn_def *sd)
{
  assert(sd);

  sd->selection_dialog = NULL;                             /* zero pointer */
  sd->caller = NULL;                                       /* zero pointer */
  sd->browse = FALSE;
}

static void
free_sctn_def(sctn_def *sd)
{
  assert(sd);
                           /* don't free GTK stuff */

  init_sctn_def(sd);
}



/* Initializes the selection.

Revision: 02.11.2003 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_init_selection(void)
{
  init_text_ring(&db_names);
  init_text_ring(&image_names);
  init_text_ring(&replay_names);
  init_sctn_def(&se);
}



/* Frees the selection.

Revision: 02.11.2003 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_free_selection(void)
{
  free_text_ring(&db_names);
  free_text_ring(&image_names);
  free_text_ring(&replay_names);
  free_sctn_def(&se);

  xplab_init_selection();
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
        - pointer to caller widget pointer
Output: - menu updated
        - widget pointer updated
*/
void
xplab_selection_menu(GtkWidget *file_menu, GtkWidget **caller)
{
  GtkWidget *file_menu_sctn;

  file_menu_sctn = gtk_menu_item_new_with_label(XPLAB_SELECTION_TITLE_ITEM XPLAB_MENU_EXT);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_sctn);

  g_signal_connect(G_OBJECT(file_menu_sctn), "activate", 
		   G_CALLBACK(cb_selection_dialog), (gpointer)caller);
}



/* Starts selection window.

Revision: 10.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_selection_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean raised;

  assert(user_data);
  assert(menuitem);
  assert(user_data);

  raised = xplab_window_raise(&se.selection_dialog, 1);

  if (raised == FALSE) {
    free_sctn_def(&se);                                   /* call repeated */
    se.caller = (GtkWidget **)user_data;            /* last succeeded call */
    xplab_selection();                          /* create selection dialog */
  }
}



/* Creates a dialog for file selection.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - selection dialog
*/
static void
xplab_selection(void)
{
  GtkWidget *slabel = NULL;
  gint      response;
  char      *err_str;
  gboolean  browse = FALSE, wait = FALSE;

  /* set the selection dialog widget                                         */
  se.selection_dialog =
    gtk_file_chooser_dialog_new(XPLAB_SELECTION_TITLE_ITEM, NULL, 
				GTK_FILE_CHOOSER_ACTION_OPEN,
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

  selection_filter(GTK_FILE_CHOOSER(se.selection_dialog));
  selection_buttons(GTK_FILE_CHOOSER(se.selection_dialog));

  do {
    wait = FALSE;

    /* wait until a response is generated through a selection button press   */
    response = gtk_dialog_run(GTK_DIALOG(se.selection_dialog));

    xplab_label_set_text(slabel, (browse == TRUE) ? "browsing..." : "");

    switch (response) {
    case GTK_RESPONSE_CANCEL: 
    case GTK_RESPONSE_DELETE_EVENT:
      browse = FALSE;

      break;

    case XPLAB_RESPONSE_BROWSE: 
      if (browse == FALSE) {
	browse = TRUE;                               /* don't destroy dialog */

	if (slabel == NULL) {                /* no label in dialog currently */
	  slabel = gtk_label_new("");
	  gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(se.selection_dialog), 
					    slabel);
	}

	xplab_label_set_text(slabel, "browsing...");
      }
      else {                                                 /* toggle state */
	browse = FALSE;                           /* destroy dialog...       */
	wait = TRUE;                              /* ...after next selection */

	gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(se.selection_dialog), 
					  NULL);
	slabel = NULL;                                    /* label destroyed */
      }

      break;

    case GTK_RESPONSE_ACCEPT: 
      {
	char *filename;

	filename = 
	  gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(se.selection_dialog));

	err_str = selection_load_file(filename);
	if (err_str) {                                     /* error occurred */
	  wait = TRUE;                                        /* keep dialog */

	  if (slabel == NULL) {              /* no label in dialog currently */
	    slabel = gtk_label_new("");                      /* create label */
	    gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(se.selection_dialog), slabel);
	  }

	  xplab_label_set_text(slabel, err_str);       /* print error string */
	  FREEMEM(err_str);
	}
	else if (browse == TRUE) {        /* no error - print default string */
	  xplab_label_set_text(slabel, "browsing...");
	}                                   /* else dialog will be destroyed */

	g_free(filename);
      } 

      break;

    default: break;
    }
  } while ((browse == TRUE) || (wait == TRUE));

  /* destroy dialog raising caller                                           */
  xplab_window_raise(se.caller, xctrl.cfg.raise);
  xplab_widget_destroy(&se.selection_dialog);

  free_sctn_def(&se);                                       /* clear pointer */
}



/* Creates a filter for the file chooser.

Revision: 17.01.2006 E.Sack

Input : - pointer to file chooser
Output: - file chooser updated
*/
static void
selection_filter(GtkFileChooser *selection_dialog)
{
  GtkFileFilter *filter;

  filter = gtk_file_filter_new();
  gtk_file_filter_add_pattern(filter, "*");
  gtk_file_filter_set_name(filter, "all");
  gtk_file_chooser_add_filter(selection_dialog, filter);

  filter = gtk_file_filter_new();
  gtk_file_filter_add_custom(filter, GTK_FILE_FILTER_FILENAME,
			     filter_db, NULL, NULL);
  gtk_file_filter_set_name(filter, "database");
  gtk_file_chooser_add_filter(selection_dialog, filter);

  filter = gtk_file_filter_new();
  gtk_file_filter_add_custom(filter, GTK_FILE_FILTER_FILENAME,
			     filter_image, NULL, NULL);
  gtk_file_filter_set_name(filter, "image");
  gtk_file_chooser_add_filter(selection_dialog, filter);

  filter = gtk_file_filter_new();
  gtk_file_filter_add_custom(filter, GTK_FILE_FILTER_FILENAME,
			     (GtkFileFilterFunc)filter_session, NULL, NULL);
  gtk_file_filter_set_name(filter, "session");
  gtk_file_chooser_add_filter(selection_dialog, filter);

  filter = gtk_file_filter_new();
  gtk_file_filter_add_custom(filter, GTK_FILE_FILTER_FILENAME,
			     filter_replay, NULL, NULL);
  gtk_file_filter_set_name(filter, "replay");
  gtk_file_chooser_add_filter(selection_dialog, filter);
}



/* Creates a custom DB filter for the file filter.

Revision: 05.08.2005 E.Sack

Input : - pointer to file filter info
        - pointer to user data
Output: - flag (file opened)
        - file chooser updated
*/
static gboolean
filter_db(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN)
{
  int      d;
  gboolean file_ok = FALSE;
  FILE     *file;
  struct stat stat_buf;

  STSFY_GCC(data);

  d = g_stat(filter_info->filename, &stat_buf);

  if (d || S_ISREG(stat_buf.st_mode) == FALSE) {
    return FALSE;                 /* we are only interested in regular files */
  }

  file = xplab_open_file(filter_info->filename, "r");

  if (file == NULL) {
    file_ok = FALSE;                                     /* cannot open file */
  }
  else {                                              /* check file for a DB */
    d = selection_scan_db(filter_info->filename, file);
    if (d) {
      file_ok = TRUE;
    }
  }

  xplab_close_file(file, filter_info->filename);

  return file_ok;
}



/* Creates a custom image filter for the file filter.

Revision: 05.08.2005 E.Sack

Input : - pointer to file filter info
        - pointer to user data
Output: - flag (file opened)
        - file chooser updated
*/
static gboolean
filter_image(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN)
{
  int      d;
  gboolean file_ok = FALSE;
  FILE     *file;
  struct stat stat_buf;

  STSFY_GCC(data);

  d = g_stat(filter_info->filename, &stat_buf);

  if (d || S_ISREG(stat_buf.st_mode) == FALSE) {
    return FALSE;                 /* we are only interested in regular files */
  }

  file = xplab_open_file(filter_info->filename, "r");

  if (file == NULL) {
    file_ok = FALSE;                                     /* cannot open file */
  }
  else {                                          /* check file for an image */
    d = selection_scan_image(file);
    if (d) {
      file_ok = TRUE;
    }
  }

  xplab_close_file(file, filter_info->filename);

  return file_ok;
}



/* Creates a custom session filter for the file filter.

Revision: 17.01.2006 E.Sack

Input : - pointer to file filter info
        - pointer to user data
Output: - flag (file opened)
        - file chooser updated
*/
static gboolean
filter_session(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN)
{
  int      d;
  gboolean file_ok = FALSE;
  FILE     *file;
  struct stat stat_buf;

  STSFY_GCC(data);

  d = g_stat(filter_info->filename, &stat_buf);

  if (d || S_ISREG(stat_buf.st_mode) == FALSE) {
    return FALSE;                 /* we are only interested in regular files */
  }

  file = xplab_open_file(filter_info->filename, "r");

  if (file == NULL) {
    file_ok = FALSE;                                     /* cannot open file */
  }
  else {
    d = selection_scan_session(file);           /* check file for an session */
    if (d) {
      file_ok = TRUE;
    }
  }

  xplab_close_file(file, filter_info->filename);

  return file_ok;
}



/* Creates a custom replay filter for the file filter.

Revision: 05.08.2005 E.Sack

Input : - pointer to file filter info
        - pointer to user data
Output: - flag (file opened)
        - file chooser updated
*/
static gboolean
filter_replay(GTK_CONVENTION_GtkFileFilterFunc_RTS_BOOLEAN)
{
  int      d;
  gboolean file_ok = FALSE;
  struct stat stat_buf;
  FILE     *file;

  STSFY_GCC(data);

  d = g_stat(filter_info->filename, &stat_buf);

  if (d || S_ISREG(stat_buf.st_mode) == FALSE) {
    return FALSE;                 /* we are only interested in regular files */
  }

  file = xplab_open_file(filter_info->filename, "r");

  if (file == NULL) {
    file_ok = FALSE;                                     /* cannot open file */
  }
  else {
    d = selection_scan_replay(filter_info->filename, file);
    if (d) {                                      /* check file for a replay */
      file_ok = TRUE;
    }
  }

  xplab_close_file(file, filter_info->filename);

  return file_ok;
}



/* Puts buttons into the selection dialog.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - selection dialog updated
*/
static void
selection_buttons(GtkFileChooser *selection_dialog)
{
  if (GTK_IS_FILE_CHOOSER_DIALOG(selection_dialog) == FALSE) {
    return;                                /* selection dialog not available */
  }

  /* put a button into the dialog */
  gtk_dialog_add_button(GTK_DIALOG(selection_dialog), "browse", 
			XPLAB_RESPONSE_BROWSE);
}



/* Scans a file for a recognition format which can be replayed.

Revision: 08.11.2003 E.Sack

Input : - pointer to name string
        - text array
Output: - text array updated
*/
static void
put_file_name(const char *source, text_ring *dest)
{
  int err;

  err = check_name(source);

  if (err) {
    return;
  }

  if (dest->td.cnt >= XPLAB_SELECTION_MAX_HISTORY_FILES && 
      XPLAB_SELECTION_MAX_HISTORY_FILES > 0) {
    dest->idx = 
      (dest->idx >= XPLAB_SELECTION_MAX_HISTORY_FILES-1) ? 0 : dest->idx+1;

    FREEMEM(dest->td.tl[dest->idx]);                       /* init pointer */
  }
  else {
    dest->idx = dest->td.cnt++;

    if (dest->idx) {
      REALLOC(char*, dest->td.cnt, dest->td.tl);
    }
    else {
      CALLOC(char*, dest->td.cnt, dest->td.tl);
    }

    dest->td.tl[dest->idx] = NULL;                         /* init pointer */
  }

  add_string(source, &dest->td.tl[dest->idx]);             /* latest entry */
}



/* Scans a file for a recognition format which can be replayed.

Revision: 07.11.2003 E.Sack

Input : - pointer to name string
Output: - flag (0: no replay file, 1: replay file)
*/
static int
selection_scan_replay(const char *replay_name, FILE *file)
{
  int      d, is_replay = 0, mborder, nborder, mweight, thold;
  char     *ident = NULL, *db_name = NULL, *filename = NULL, *options = NULL;
  rb_lst   rblst;
  cand_lst hit_lst; 

  assert(replay_name);
  assert(file);

  init_rb_lst(&rblst);
  init_cand_lst(&hit_lst); 

  d = check_name(replay_name);
  assert(!d);

  d = xplab_read_recognition_file(replay_name, file, &rblst, &db_name, 
				  &ident, &filename, &mborder, &nborder, 
				  &mweight, &thold, &hit_lst, &options, 0);

  free_rb_lst(&rblst);
  free_cand_lst(&hit_lst); 
  FREEMEM(ident);
  FREEMEM(db_name);
  FREEMEM(filename);
  FREEMEM(options);

  if (!d) {
    is_replay = 1;
  }

  return is_replay;
}



/* Scans a file for a DB.

Revision: 02.11.2003 E.Sack

Input : - pointer to DB name string
Output: - flag (0: no DB file, 1: DB file)
*/
static int
selection_scan_db(const char *db_name, FILE *file)
{
  int    d, is_db = 0;
  char   *version_ctrl = NULL;
  db_def db;

  assert(file);

  d = check_name(db_name);
  assert(!d);

  db = xplab_read_db_file(db_name, file, &version_ctrl, 0);

  d = check_name(db.db_fname);

  if (!d) {
    is_db = 1;
  }

  free_db_def(&db);
  FREEMEM(version_ctrl);

  return is_db;
}



/* Scans a file for a PNM or session format.

Revision: 02.11.2003 E.Sack

Input : - pointer to file stream
Output: - flag (0: no PNM file, 1: PNM file)
*/
static int
selection_scan_image(FILE *file)
{
  int        mgv;
  area_def   size;
  PNM_FORMAT format;

  assert(file);

  format = read_pnm_hdr(file, &size, &mgv);

  if (format == PNM_FORMAT_UNDEF) {
    return 0;                                  /* not an extended PNM format */
  }

  return 1;
}



/* Scans a file for a session format.

Revision: 17.01.2006 E.Sack

Input : - pointer to file stream
Output: - flag (0: no session file, 1: session file)
*/
static int
selection_scan_session(FILE *file)
{
  int        mgv;
  area_def   size;
  PNM_FORMAT format;

  assert(file);

  format = read_pnm_hdr(file, &size, &mgv);

  if ((format == PNM_FORMAT_UNDEF) || 
      ((format != PXPLAB2) && (format != PXPLAB5))) {
    return 0;
  }

  return 1;
}



/* Accepts selection dialog and try to load file.

Revision: 17.01.2006 E.Sack

Input : - pointer to file spec string
Output: - pointer to error string
*/
static char *
selection_load_file(char *filespec)
{
  int      err, selection_ok = 0, err_image_lock = 0, err_db_lock = 0;
  char     *err_txt = NULL;
  FILE *file;

  err = check_name(filespec);

  if (!err) {
    file = xplab_open_file(filespec, "r");

    while (file != NULL) {
      selection_ok = selection_scan_image(file);  /* check file for an image */

      if (selection_ok) {
	int ret;
	struct stat buf;
	gchar *btmp = default_session_filename(filespec);/* create according session filename */

	ret = g_stat(btmp, &buf);

	if (!ret && strcmp(btmp, filespec)) {
	  string_to_log(main_log, "INFO: session file `", btmp, "' exists.\n");
	}

	put_file_name(filespec, &image_names);        /* put name to history */
	err_image_lock = xplab_input_image(image_names.td.tl[image_names.idx]);
	FREEMEM(btmp);
	break;
      }

      rewind(file);

      selection_ok = selection_scan_replay(filespec, file);/*check for a replay */
      if (selection_ok) {
	put_file_name(filespec, &replay_names);       /* put name to history */
	xplab_input_replay_file();                          /* renew history */
	break;
      }

      selection_ok = selection_scan_db(filespec, file); /*check file for a DB*/
      if (selection_ok) {
	put_file_name(filespec, &db_names);           /* put name to history */

	err_db_lock = xplab_input_db(db_names.td.tl[db_names.idx]);
      }

      break;
    }

    xplab_close_file(file, filespec);
  }
  else {
    add_string("no file selected", &err_txt);
    selection_ok = 1;
  }

  err = 1;

  if (selection_ok && !err_image_lock && !err_db_lock) {
    err = 0;
  }
  else if (!selection_ok) {
    add_string("WARNING: not a loadable file", &err_txt);
  }
  else if (err_image_lock) {
    add_string("WARNING: can't load new image - current image locked", &err_txt);
  }
  else if (err_db_lock) {
    add_string("WARNING: can't load new database - current database locked", 
	       &err_txt);
  }

  return err_txt;
}
