/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
******************************************************************************
Debugging support functions only (Print Outs), basically not needed for XPLAB.

The usage of the functions below beginning with Print... are primarily for
observing/debugging the data flow produced by the code. They have their
advantages in printing many infos in a structured manner without be bothered
coding the same debugging code again and again.

For example, if one would print out calculated sections in certain stages of
computation, he inserts the configurable function `Print_sector' with

  :
Print_sector("RECOG. NB SECT IS ", "\n", hlist.c[i].sec, 1);
  :

and he would possibly get a list like

  :
RECOG. NB SECT IS 123-130x345-366
RECOG. NB SECT IS 129-135x345-366
RECOG. NB SECT IS 137-155x340-360
RECOG. NB SECT IS 160-170x341-369
  :

on the screen. If the printout is unwanted temporarely, set the last 1 to a 0
in the function.

All the functions are supported with a prefix and a suffix by the user. With
a plain `printf' before and after such a function one can have well 
documenting lines.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_debug.h"
#include "xplab_tprint.h"

/* local prototypes ********************************************************/
static void sPrint_cand_lst(cand_lst cl, char **str);
static void sPrint_cand_lst_indexed(cand_lst cl, int_lst ilst, char **str);
static void sPrint_cand_def(cand_def kr, char *str);
static void sPrint_cv_def_i(cv_def_i kd, char *str);
static void sPrint_sector_def(sector_def sc, char *str);
static void sPrint_area_def(area_def ar, char *str);
static void sPrint_sign(cand_lst kl, cls_def *cls, char **str);
static void sPrint_sign_indexed(cand_lst kl, int *idx, cls_def *cls, char **str);
static void sPrint_sign_numbered(cand_lst kl, int_lst ilst, cls_def *cls, 
			  char **str);
static char *sPrint_knl(cand_lst kl);
static char *sPrint_nbr(cand_lst kl);
static char *sPrint_cv(cand_lst kl);
static void sPrint_col_def(col_def cd, char *str);
static void sPrint_int_lst(int_lst ilst, char **str);
static void sPrint_sector_lst(sector_lst sl, char **str, int width);

/* Functions ****************************************************************/

void
Print_int_lst(char *prefix, char *suffix, int_lst ilst, int on)
{
  char *int_lst_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_int_lst(ilst, &int_lst_str);

  fprintf(stdout, "%s%s%s", prefix, int_lst_str, suffix);

  FREEMEM(int_lst_str);
}

static void
sPrint_int_lst(int_lst ilst, char **str)
{
  int  i;
  char buf[TSTLEN];

  g_snprintf(buf, sizeof(buf), "int_lst.cnt: %d\n", ilst.cnt);
  add_string(buf, str);

  for (i = 0; i < ilst.cnt; i++) {
    g_snprintf(buf, sizeof(buf), "%d%c", ilst.nr[i], ((i%8)+1)==8 ? '\n' : ' ');
    add_string(buf, str);
  }

  add_string((((i-1)%8)+1)!=8 ? "\n" : " ", str);
}

void
Print_cand_def(char *prefix, char *suffix, cand_def kr, int on)
{
  char cand_def_str[TSTLEN];

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_cand_def(/*prefix, */ kr, cand_def_str);

  fprintf(stdout, "%s%s%s", prefix, cand_def_str, suffix);
}

static void
sPrint_cand_def(/*char *prefix, */ cand_def kr, char *str)
{
  char cv_def_i_m[TSTLEN], sector_roi[TSTLEN], sector_knl[TSTLEN];
  char sector_hdx[TSTLEN], sector_ext[TSTLEN];
  char *int_lst_str = 0;

  sPrint_cv_def_i(kr.m, cv_def_i_m);
  sPrint_sector_def(kr.roi, sector_roi);
  sPrint_sector_def(kr.ext, sector_ext);
  sPrint_sector_def(kr.knl, sector_knl);
  sPrint_sector_def(kr.hdx, sector_hdx);
  sPrint_int_lst(kr.crb, &int_lst_str);

  g_snprintf(str, TSTLEN, "%s  RB[%d] CLS[%d][%d] %s %s %s %s\nCRB-%s", 
	   cv_def_i_m, kr.rbn, kr.c, kr.n, sector_roi, sector_ext, sector_knl,
	   sector_hdx, int_lst_str);

  str[TSTLEN-2] = 'T';                                /* truncation */
  str[TSTLEN-1] = '\0';

  FREEMEM(int_lst_str);
}

void
Print_cand_lst(char *prefix, char *suffix, cand_lst cl, int on)
{
  char *cand_lst_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_cand_lst(cl, &cand_lst_str);

  fprintf(stdout, "%s%s%s", prefix, cand_lst_str, suffix);

  FREEMEM(cand_lst_str);
}

static void
sPrint_cand_lst(cand_lst cal, char **str)
{
  int  i;
  char buf[TSTLEN];

  g_snprintf(buf, sizeof(buf), "cand_lst.cnt: %d\n", cal.cnt);
  add_string(buf, str);

  for (i = 0; i < cal.cnt; i++) {
    sPrint_cand_def(cal.c[i], buf);
    cat_string(buf, "\n", "", str);
  }
}

void
Print_cand_lst_indexed(char *prefix, char *suffix, cand_lst cl, int_lst ilst, 
		       int on)
{
  char *cand_lst_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_cand_lst_indexed(cl, ilst, &cand_lst_str);

  fprintf(stdout, "%s%s%s", prefix, cand_lst_str, suffix);

  FREEMEM(cand_lst_str);
}

static void
sPrint_cand_lst_indexed(cand_lst cal, int_lst ilst, char **str)
{
  int  i, idx;
  char buf[TSTLEN];

  int_to_string("cand_lst.cnt: ", cal.cnt, "  ", str);
  int_to_string("int_lst.cnt: ", ilst.cnt, "\n", str);

  for (i = 0; i < ilst.cnt; i++) {
    idx = ilst.nr[i];

    assert(idx >=0);

    sPrint_cand_def(cal.c[idx], buf);
    add_string(buf, str);
  }
}

void
Print_cv_def_i(char *prefix, char *suffix, cv_def_i kdi, int on)
{
  char cv_def_i_str[TSTLEN];

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_cv_def_i(kdi, cv_def_i_str);

  fprintf(stdout, "%s%s%s", prefix, cv_def_i_str, suffix);
}

static void
sPrint_cv_def_i(cv_def_i kdi, char *str)
{
  g_snprintf(str, TSTLEN, "KW %f lag %d,%d", kdi.cv, kdi.x, kdi.y);

  str[TSTLEN-2] = 'T';                                /* truncation */
  str[TSTLEN-1] = '\0';
}

void
Print_sector_def(char *prefix, char *suffix, sector_def sc, int on)
{
  char sector_str[TSTLEN];

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_sector_def(sc, sector_str);

  fprintf(stdout, "%s%s%s", prefix, sector_str, suffix);
}

static void
sPrint_sector_def(sector_def sc, char *str)
{
  g_snprintf(str, TSTLEN, "%d-%dx%d-%d  %d x %d", sc.column.start, 
	   sc.column.end, sc.line.start, sc.line.end, 
	   INTERVAL_SIZE(sc.column), INTERVAL_SIZE(sc.line));

  str[TSTLEN-2] = 'T';                                /* truncation */
  str[TSTLEN-1] = '\0';
}

void
Print_col_def(char *prefix, char *suffix, col_def cd, int on)
{
  char col_def_str[TSTLEN];

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_col_def(cd, col_def_str);

  fprintf(stdout, "%s%s%s", prefix, col_def_str, suffix);
}

static void
sPrint_col_def(col_def cd, char *str)
{
  g_snprintf(str, TSTLEN, "(%d,%d,%d)", cd.r, cd.g, cd.b);

  str[TSTLEN-2] = 'T';                                /* truncation */
  str[TSTLEN-1] = '\0';
}

void
Print_area_def(char *prefix, char *suffix, area_def ar, int on)
{
  char area_def_str[TSTLEN];

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_area_def(ar, area_def_str);

  fprintf(stdout, "%s%s%s", prefix, area_def_str, suffix);
}

static void
sPrint_area_def(area_def ar, char *str)
{
  g_snprintf(str, TSTLEN, "%dx%d", ar.width, ar.height);

  str[TSTLEN-2] = 'T';                                /* truncation */
  str[TSTLEN-1] = '\0';
}

void
Print_knl(char *prefix, char *suffix, cand_lst kl, int on)
{
  char *knl_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  knl_str = sPrint_knl(kl);

  fprintf(stdout, "%s%s%s", prefix, knl_str, suffix);

  FREEMEM(knl_str);
}

static char *
sPrint_knl(cand_lst kl)
{
  int  i, len;
  char *knl_str = 0;
  char sector_str[TSTLEN], buf[TSTLEN];

  if (!kl.cnt) {
    add_string(EMPTY_STR, &knl_str);
    return knl_str;
  }

  for (i = kl.cnt-1; i >= 0 ; i--) {           /* list build from the right */
    sPrint_sector_def(kl.c[i].knl, sector_str);

    if (knl_str) {
      len = strlen(sector_str) + strlen(knl_str) + TSTLEN;
      REALLOC(char, len, knl_str);
    }
    else {
      len = strlen(sector_str) + TSTLEN;
      CALLOC(char, len, knl_str);
    }

    g_snprintf(buf, sizeof(buf), "%d(%s) ", i, sector_str);
    strcat(knl_str, buf);
  }

  return knl_str;
}

void
Print_nbr(char *prefix, char *suffix, cand_lst kl, int on)
{
  char *nbr_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  nbr_str = sPrint_nbr(kl);

  fprintf(stdout, "%s%s%s", prefix, nbr_str, suffix);

  FREEMEM(nbr_str);
}

static char *
sPrint_nbr(cand_lst kl)
{
  int  i, len;
  char *nbr_str = 0;
  char buf[TSTLEN];

  if (!kl.cnt) {
    add_string(EMPTY_STR, &nbr_str);
    return nbr_str;
  }

  for (i = kl.cnt-1; i >= 0 ; i--) {           /* list build from the right */
    if (nbr_str) {
      len = strlen(nbr_str) + TSTLEN;
      REALLOC(char, len, nbr_str);
    }
    else {
      len = TSTLEN;
      CALLOC(char, len, nbr_str);
    }

    g_snprintf(buf, sizeof(buf), "%d(%d,%d) ", i, kl.c[i].c, kl.c[i].n);
    strcat(nbr_str, buf);
  }

  return nbr_str;
}

void
Print_cv(char *prefix, char *suffix, cand_lst kl, int on)
{
  char *cv_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  cv_str = sPrint_cv(kl);

  fprintf(stdout, "%s%s%s", prefix, cv_str, suffix);

  FREEMEM(cv_str);
}

static char *
sPrint_cv(cand_lst kl)
{
  int  i, len;
  char *cv_str = 0;
  char buf[TSTLEN];

  if (!kl.cnt) {
    add_string(EMPTY_STR, &cv_str);
    return cv_str;
  }

  for (i = kl.cnt-1; i >= 0 ; i--) {           /* list build from the right */
    if (cv_str) {
      len = strlen(cv_str) + TSTLEN;
      REALLOC(char, len, cv_str);
    }
    else {
      len = TSTLEN;
      CALLOC(char, len, cv_str);
    }

    g_snprintf(buf, sizeof(buf), "%d(%.3f) ", i, kl.c[i].m.cv);
    strcat(cv_str, buf);
  }

  return cv_str;
}

void
Print_sign(char *prefix, char *suffix, cand_lst kl, cls_def *cls, int on)
{
  char *string = 0;

  if (!on) {
    return;
  }
  else if (on == -1 && !kl.cnt) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_sign(kl, cls, &string);

  fprintf(stdout, "%s%s%s", prefix, string, suffix);

  FREEMEM(string);
}

static void
sPrint_sign(cand_lst kl, cls_def *cls, char **str)
{
  int i;

  if (!kl.cnt) {
    add_string(EMPTY_STR, str);
    return;
  }

  for (i = kl.cnt-1; i >= 0 ; i--) {           /* list build from the right */
    add_string(cls[kl.c[i].c].att.sign, str);
  }
}

void
Print_sign_indexed(char *prefix, char *suffix, cand_lst kl, int *idx, 
		   cls_def *cls, int on)
{
  char *string = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_sign_indexed(kl, idx, cls, &string);

  fprintf(stdout, "%s%s%s", prefix, string, suffix);

  FREEMEM(string);
}

static void
sPrint_sign_indexed(cand_lst kl, int *idx, cls_def *cls, char **str)
{
  int empty, i;

  if (!kl.cnt) {
    add_string(EMPTY_STR, str);
    return;
  }

  for (i = kl.cnt-1, empty = 1; i >= 0 ; i--) {/* list build from the right */
    if (idx[i] == BAD_STRING) {
      continue;
    }

    empty = 0;

    add_string(cls[kl.c[i].c].att.sign, str);
  }

  if (empty) {
    add_string(EMPTY_STR, str);
  }
}

void
Print_sign_numbered(char *prefix, char *suffix, cand_lst kl, int_lst ilst, 
		   cls_def *cls, int on)
{
  char *sign_numbered_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_sign_numbered(kl, ilst, cls, &sign_numbered_str);

  fprintf(stdout, "%s%s%s", prefix, sign_numbered_str, suffix);

  FREEMEM(sign_numbered_str);
}

static void
sPrint_sign_numbered(cand_lst kl, int_lst ilst, cls_def *cls, char **str)
{
  int i;

  if (!kl.cnt || !ilst.cnt) {
    add_string(EMPTY_STR, str);
    return;
  }

  for (i = 0; i < ilst.cnt; i++) {
    add_string(cls[kl.c[ilst.nr[i]].c].att.sign, str);
  }
}

void
Test_sector(char *prefix, char *suffix, sector_def sc, int on)
{
  int  flaw;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  flaw = sc.column.end < sc.column.start || sc.line.end < sc.line.start;

  if (flaw) {
    Print_sector_def(prefix, suffix, sc, on);
  }
}

void
Print_sector_lst(char *prefix, char *suffix, sector_lst sl, int width, int on)
{
  char *sector_lst_str = 0;

  if (!on) {
    return;
  }
  else {
    P_print_debug = 1;
  }

  sPrint_sector_lst(sl, &sector_lst_str, width);

  fprintf(stdout, "%s%s%s", prefix, sector_lst_str, suffix);

  FREEMEM(sector_lst_str);
}

static void
sPrint_sector_lst(sector_lst sl, char **str, int width)
{
  int  i, cnt = width;
  char buf[TSTLEN];

  g_snprintf(buf, sizeof(buf), "sector_lst.cnt: %d\n", sl.cnt);
  add_string(buf, str);

  for (i = 0; i < sl.cnt; i++) {
    sPrint_sector_def(sl.sdl[i], buf);
    add_string(buf, str);

    cnt--;

    if (cnt <= 0) {
      add_string("\n", str);
      cnt = width;
    }
    else {
      add_string(" ", str);
    }
  }
}

