/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <gtk/gtk.h>
#include <math.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_training_be.h"
#include "xplab_tprint.h"
#include "xplab_maintenance_be.h"
#include "xplab_algorithm.h"

/* local definitions *******************************************************/
#define CLASS_ATTRIBUTES_DIALOG             "class attributes dialog"
#define CLASS_ATTRIBUTES_DIALOG_CLASS       "CLASS:"
#define CLASS_ATTRIBUTES_DIALOG_TRANSLATION "TRANSLATION:"
#define CLASS_ATTRIBUTES_DIALOG_COMMENT     "COMMENT:"
#define CLASS_ATTRIBUTES_DIALOG_ALIGNMENT   "ALIGNMENT:"
#define CLASS_ATTRIBUTES_DIALOG_PRINTABLE   "PRINTABLE:"
#define CLASS_ATTRIBUTES_DIALOG_FIXED       "FIXED:"

/* Functions assembling widgets */
static GtkWidget *xplab_combo_text(GtkWidget **combo, gboolean entry_is_editable, glist_def glist, guint16 entry_max_length, GtkSignalFunc entry_cb, gpointer entry_cb_data);
static GtkWidget *combo_list_empty(entry_def *comment, GtkSignalFunc entry_cb, guint16 entry_max_length);
static GtkWidget *combo_list_boolean(entry_def *fixed);
static GtkWidget *combo_list_printable(entry_def *printable);
static GtkWidget *combo_list_alignment(entry_def *alignment);
static void zoom_scroll_window(GtkWidget *vbox, scroll_def *swin, area_def drawing_size, GtkSignalFunc mouse_cb, GtkSignalFunc slider_cb, GtkSignalFunc expose_cb);
static gfloat net_displayed(gfloat start, gfloat visible, int allocation, int offset);
static void magnitude_menu_item(gfloat mag, gfloat mag_act, GSList **group, GtkWidget *magnitude_menu, GtkSignalFunc magnitude_cb, gpointer udata);
static void magnitude_content(col_def *src_img, area_def src_size, gfloat src_start_x, gfloat src_start_y, gfloat magnitude, pic_def *dst_img, gfloat visible_x, gfloat visible_y, int allocation_x, int allocation_y, gint force);
static void scale_up(col_def *src_img, area_def src_size, gfloat src_start_x, gfloat src_start_y, gfloat magnitude, pic_def dst_img, gfloat visible_x, gfloat visible_y, int allocation_x, int allocation_y);

/* Handlers/Callbacks */
static void cb_class_text_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_translation_text_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_comment_text_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);

/* Functions ***************************************************************/

/* Calculation of the biggest rectangular frame of marked pixels in a sector. 

Input : - marking
        - pointer to sector
        - image
        - flag (0: exakt marking, 1: all markings but background)
Output: - sector updated
        - flag (0: no marking found, 1: marking found)
Precondition: - marking ignored if input flag is 1
Postcondition: - maximal size of updated sector is input sector size
*/
int 
contour_frame(PICT mark, sector_def *fr, pic_def pic, int all)
{
  int z, s, zs, ze, ss, se, zz, found = 0, px;

  zs = fr->line.end;                                       /* input sector */
  ze = fr->line.start;
  ss = fr->column.end;
  se = fr->column.start;

  for (z = fr->line.start; z <= fr->line.end; z++) {
    zz = z*pic.ar.width;

    for (s = fr->column.start; s <= fr->column.end; s++) {
      px = pic.mem[s + zz];

      if ((px == mark && !all) || (px != BACKGROUND && all)) {
	zs = XPMIN(zs, z);
	ze = XPMAX(ze, z);
	ss = XPMIN(ss, s);
	se = XPMAX(se, s);
	found = 1;
      }
    }
  }        

  fr->line.start  = zs;      /* input sector overwritten with marking size */
  fr->line.end    = ze;
  fr->column.start = ss;
  fr->column.end   = se;

  return found;
}



/* Checks the training resources.

Revision: 25.10.2003 E.Sack

Input : - DB
        - pointer to TPS ID
        - pointer to image file name string
        - image
        - RB list
Output: - TPS index (-1: error)
        - message output
*/
int
check_training_resources(db_def db, char *tps_id, char *filename, 
			 pic_def rpic, rb_lst rbl)
{
  int        err, tps_idx;
  char       *id = NULL;
  DATA_STATE is_state;

  err = check_name(filename);

  if (AREA_SIZE(rpic.ar) == 0) {
    xplab_text_insert(main_log, TRAINING_WARNING_1);
    return -1;
  }

  if (err) {
    xplab_text_insert(main_log, TRAINING_WARNING_2);
    return -1;
  }

  if (rpic.mem == NULL) {
    xplab_text_insert(main_log, TRAINING_WARNING_3);
    return -1;
  }

  if (rpic.col) {
    xplab_text_insert(main_log, TRAINING_WARNING_4);
    return -1;
  }

  if (rbl.cnt == 0) {
    xplab_text_insert(main_log, TRAINING_WARNING_5);
    return -1;
  }

  is_state = state_db_def(db);

  if (is_state == STATE_IS_INIT) {
    return -1;                                 /* message output elsewhere */
  }

  err = check_name(db.db_fname);

  if (err) {
    xplab_text_insert(main_log, TRAINING_WARNING_6);
    return -1;
  }

  if (db.tpl.cnt == 0) {                /* at least one empty TPS required */
    xplab_text_insert(main_log, TRAINING_WARNING_7);
    return -1;
  }

  err = check_name(tps_id);

  if (err) {
    xplab_text_insert(main_log, TRAINING_WARNING_8);
    return -1;
  }

  add_string(tps_id, &id);
  tps_idx = search_tps_id(&id, db.tpl);
  err = strcmp(id, tps_id);
  FREEMEM(id);

  if (err) {
    xplab_text_insert(main_log, TRAINING_WARNING_9);
    return -1;
  }

  return tps_idx;
}



/* Assembles the NB attributes and pattern into a class list, which is
   extended accordingly.

Revision: 29.07.2005 E.Sack

Input : - class attribute entrys
Output: - class
*/
cls_att
xget_entry_cls_att(cls_att_entry att)
{
  int      flag, len;
  gchar    *buf;
  cls_att  att_tmp;
  GtkEntry *centry;

  init_cls_att(&att_tmp);

  assert(GTK_IS_COMBO_BOX(att.class.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.class.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */
  len = strlen(buf);
  assert(len < MAX_LEN_CLASS_NAME);         /* entry must be parsed before */
  add_string(buf, &att_tmp.class);

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.translatn.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.translatn.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */
  len = strlen(buf);
  assert(len < MAX_LEN_CLASS_TRANS);        /* entry must be parsed before */
  add_string(buf, &att_tmp.sign);

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.commnt.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.commnt.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */
  len = strlen(buf);
  assert(len < MAX_LEN_CLASS_COMMENT);      /* entry must be parsed before */
  add_string(buf, &att_tmp.crem);

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.alignment.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.alignment.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = string_to_aligmnt(buf, &att_tmp.alignment);
  assert(flag);                             /* entry must be parsed before */

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.printable.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.printable.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  if (!strcmp(buf, "TRUE")) {
    att_tmp.printable = 1;
  }
  else if (!strcmp(buf, "FALSE")) {
    att_tmp.printable = 0;
  }
  else {
    assert(0);
  }

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.fixed.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.fixed.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  if (!strcmp(buf, "TRUE")) {
    att_tmp.fixed = 1;
  }
  else if (!strcmp(buf, "FALSE")) {
    att_tmp.fixed = 0;
  }
  else {
    assert(0);
  }

  FREEMEM(buf);

  return att_tmp;
}



/* Wraps around an index to the next logical value.

Revision: 28.08.2003 E.Sack

Input : - pointer to index
        - number of indizes
Output: - index updated
Postcondition: - index between 0 and number-1
               - index unchanged if (0 <= index < number) is true.
*/
void
xwrap_idx(int *idx, int number)
{
  assert(number > 0);

  if (*idx >= number) {
    *idx = 0;
  }
  else if (*idx < 0) {
    *idx = number-1;
  }
}



/* Clears text entry.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_clear_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  entry_def *ety = (entry_def *)user_data;
  GtkEntry  *centry;

  assert(button);
  assert(user_data);

  ety = (entry_def *)user_data;
  assert(GTK_IS_COMBO_BOX(ety->combo));

  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(ety->combo))->child);

  xplab_entry_set_text(centry, "");                         /* clear entry */
}



/* Creates an attribute input mask.

Revision: 29.07.2005 E.Sack

Input : - pointer to container frame
        - pointer to class attributes setting
        - pointers to attribute signal handlers
Output: - container box updated
Precondition: - signal handlers may be NULL
*/
void
edit_class_attributes(GtkWidget *frame, cls_att_entry *att)
{
  GtkWidget *vbox, *table, *class_box, *label_fixed, *label_printable;
  GtkWidget *label_class, *label_translation, *label_comment, *fixed_box;
  GtkWidget *translation_box, *comment_box, *alignment_box, *printable_box;
  GtkWidget *label_alignment;

  assert(frame);
  assert(att);
  assert(GTK_IS_FRAME(frame) == TRUE);

  gtk_frame_set_label(GTK_FRAME(frame), CLASS_ATTRIBUTES_DIALOG);

  vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  label_class = gtk_label_new(CLASS_ATTRIBUTES_DIALOG_CLASS);
  class_box = combo_list_empty(&att->class, 
			       GTK_SIGNAL_FUNC(cb_class_text_entry), 
			       MAX_LEN_CLASS_NAME);

  label_translation = gtk_label_new(CLASS_ATTRIBUTES_DIALOG_TRANSLATION);
  translation_box = combo_list_empty(&att->translatn, 
				     GTK_SIGNAL_FUNC(cb_translation_text_entry),
				     MAX_LEN_CLASS_TRANS);

  label_comment = gtk_label_new(CLASS_ATTRIBUTES_DIALOG_COMMENT);
  comment_box = combo_list_empty(&att->commnt,
				 GTK_SIGNAL_FUNC(cb_comment_text_entry), 
				 MAX_LEN_CLASS_COMMENT);

  label_alignment = gtk_label_new(CLASS_ATTRIBUTES_DIALOG_ALIGNMENT);
  alignment_box = combo_list_alignment(&att->alignment);

  label_printable = gtk_label_new(CLASS_ATTRIBUTES_DIALOG_PRINTABLE);
  printable_box = combo_list_printable(&att->printable);

  label_fixed = gtk_label_new(CLASS_ATTRIBUTES_DIALOG_FIXED);
  fixed_box = combo_list_boolean(&att->fixed);

  att->infoline = gtk_label_new (NULL);

  table = gtk_table_new(8, 3, FALSE);
  gtk_container_add(GTK_CONTAINER(vbox), table);

  gtk_table_attach_defaults(GTK_TABLE(table), label_class,       0, 1, 0, 1);
  gtk_table_attach_defaults(GTK_TABLE(table), label_translation, 0, 1, 1, 2);
  gtk_table_attach_defaults(GTK_TABLE(table), label_comment,     0, 1, 2, 3);
  gtk_table_attach_defaults(GTK_TABLE(table), label_alignment,   0, 1, 3, 4);
  gtk_table_attach_defaults(GTK_TABLE(table), label_printable,   0, 1, 4, 5);
  gtk_table_attach_defaults(GTK_TABLE(table), label_fixed,       0, 1, 5, 6);

  gtk_table_attach_defaults(GTK_TABLE(table), class_box,       1, 2, 0, 1);
  gtk_table_attach_defaults(GTK_TABLE(table), translation_box, 1, 2, 1, 2);
  gtk_table_attach_defaults(GTK_TABLE(table), comment_box,     1, 2, 2, 3);
  gtk_table_attach_defaults(GTK_TABLE(table), alignment_box,   1, 2, 3, 4);
  gtk_table_attach_defaults(GTK_TABLE(table), printable_box,   1, 2, 4, 5);
  gtk_table_attach_defaults(GTK_TABLE(table), fixed_box,       1, 2, 5, 6);

  gtk_table_attach_defaults(GTK_TABLE(table), att->infoline,    0, 2, 7, 8);

  gtk_table_set_row_spacing(GTK_TABLE(table), 5, 5);
  gtk_table_set_row_spacing(GTK_TABLE(table), 6, 5);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
}



/* Creates a combo entry widget.

Revision: 23.06.2004 E.Sack

Input : - pointer to entry definition
Output: - pointer to container
*/
static GtkWidget *
combo_list_printable(entry_def *printable)
{
  GtkWidget *box;

  assert(printable);

  box = combo_list_boolean(printable);

  return box;
}



/* Creates a combo entry widget.

Revision: 23.06.2004 E.Sack

Input : - pointer to entry definition
Output: - pointer to container
*/
static GtkWidget *
combo_list_boolean(entry_def *fixed)
{
  GtkWidget *box;

  assert(fixed);
  assert(fixed->elist.items == NULL);                  /* allocation control */
  assert(fixed->elist.fail == NULL);

  fixed->elist.items = g_list_append(fixed->elist.items, "FALSE");
  fixed->elist.items = g_list_append(fixed->elist.items, "TRUE");
  add_string(CTRL_FLOW_2, &fixed->elist.fail);

  box = xplab_combo_text(&fixed->combo, FALSE, fixed->elist, 
			 XPLAB_CLASS_ENTRY_LENGTH, NULL, NULL);

  free_glist_def(&fixed->elist);

  return box;
}



/* Creates a combo entry widget.

Revision: 29.07.2005 E.Sack

Input : - pointer to entry definition
        - entry handler
        - text entry length
Output: - pointer to container
*/
static GtkWidget *
combo_list_empty(entry_def *cent, GtkSignalFunc entry_cb, 
		 guint16 entry_max_length)
{
  GtkWidget *box;

  assert(cent);

  cent->elist.items = NULL;             /* the empty list is initialized here */

  box = xplab_combo_text(&cent->combo, TRUE, cent->elist, 
			 entry_max_length, entry_cb, (gpointer)cent);

  xset_clicked_button(box, GTK_SIGNAL_FUNC(cb_clear_button), "C", NULL, NULL,
		      (gpointer)cent);

  return box;
}



/* Creates a combo entry widget.

Revision: 29.07.2005 E.Sack

Input : - pointer to entry definition
        - entry handler
        - text entry length
        - handler left
        - handler right
        - handler clear
Output: - pointer to container
*/
GtkWidget *
combo_list_find(entry_def *cent, guint16 entry_max_length, 
		GtkSignalFunc combo_cb, gpointer combo_cb_data)
{
  GtkWidget *box;

  assert(cent);

  box = xplab_combo_text(&cent->combo, FALSE, cent->elist,
			 entry_max_length, NULL, NULL);

  gtk_combo_box_set_active(GTK_COMBO_BOX(cent->combo), 0); /* default column*/

  g_signal_connect(G_OBJECT(cent->combo), "changed",
		   G_CALLBACK(combo_cb), combo_cb_data);

  return box;
}



/* Creates a combo entry widget.

Revision: 23.06.2004 E.Sack

Input : - pointer to entry definition
Output: - pointer to container
*/
static GtkWidget *
combo_list_alignment(entry_def *alignment)
{
  GtkWidget *box;

  assert(alignment);
  assert(alignment->elist.items == NULL);               /* allocation control */
  assert(alignment->elist.fail == NULL);

  alignment->elist.items = g_list_append(alignment->elist.items, "UPPERLINE");
  alignment->elist.items = g_list_append(alignment->elist.items, "MIDDLELINE");
  alignment->elist.items = g_list_append(alignment->elist.items, "BASELINE");
  alignment->elist.items = g_list_append(alignment->elist.items, "UNDERLINE");
  add_string(CTRL_FLOW_3, &alignment->elist.fail);

  box = xplab_combo_text(&alignment->combo, FALSE, alignment->elist, 
			 XPLAB_CLASS_ENTRY_LENGTH, NULL, NULL);

  free_glist_def(&alignment->elist);

  return box;
}



/* Checks the user class name input.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_class_text_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  entry_def *class = (entry_def *)user_data;

  assert(entry);
  assert(user_data);

  xplab_check_entry(*class, MAX_LEN_CLASS_NAME);

  put_history_to_combo_store(class);
}



/* Checks the user translation string input.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_translation_text_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  entry_def *translation = (entry_def *)user_data;

  assert(entry);
  assert(user_data);

  xplab_check_entry(*translation, MAX_LEN_CLASS_TRANS);

  put_history_to_combo_store(translation);
}



/* Callback for the user comment string input.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_comment_text_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  entry_def *comment = (entry_def *)user_data;

  assert(entry);
  assert(user_data);

  xplab_check_entry(*comment, MAX_LEN_CLASS_COMMENT);

  put_history_to_combo_store(comment);
}



/* Checks the user comment string input.

Revision: 21.06.2003 E.Sack

Input : - pointer to training settings
Output: - flag
*/
int 
xplab_check_entry(entry_def cent, guint16 entry_max_length)
{
  int      flag;
  gchar    *buf, *ret_buf;
  GtkEntry *centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(cent.combo))->child);

  buf = xplab_entry_get_text(centry);                       /* fetch string */

  ret_buf = xplab_parse_string(buf, &flag, entry_max_length); /* correction */
  FREEMEM(buf);

  xplab_entry_set_text(centry, ret_buf);                       /* redisplay */
  FREEMEM(ret_buf);

  buf = xplab_string_message(flag);                         /* assemble msg */

  xplab_label_set_text(*cent.info_pt, buf);                  /* display msg */

  FREEMEM(buf);

  flag = xplab_string_usage(flag);

  return flag;
}



/* Sets reset store.

Revision: 06.08.2005 E.Sack

Input : - pointer to class attributes
Output: - attribute stores updated
*/
void
update_history_cls_att_entry(cls_att_entry *att)
{
  put_history_to_combo_store(&att->class);
  put_history_to_combo_store(&att->translatn);
  put_history_to_combo_store(&att->commnt);
}



/* Wrapper for gtk_combo_box_append_text.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
void 
combo_box_append_text(GTK_CONVENTION_GFunc_RTS_VOID)
{
  assert(GTK_IS_COMBO_BOX(user_data));
  gtk_combo_box_append_text(GTK_COMBO_BOX(user_data), (gchar *)data);
}



/* Creates a general combo entry widget.

Revision: 29.07.2005 E.Sack

Input : - pointer to text entry widget pointer
        - flag (FALSE: text entry is not editable, TRUE: text entry is editable)
        - pointer to GList for combo widget
        - pointer to default text entry text string
        - text entry length
        - text entry handler
        - pointer to text entry handler data
Output: - pointer to container of combo entry widget
        - pointer to text entry widget updated
*/
static GtkWidget *
xplab_combo_text(GtkWidget **combo, gboolean entry_is_editable, glist_def glist, 
		 guint16 entry_max_length, GtkSignalFunc entry_cb, 
		 gpointer entry_cb_data)
{
  GtkWidget *hbox;
  GtkEntry  *centry;

  assert(combo);
  assert(entry_max_length>1);

  entry_max_length--;                /* pedantic: string terminating \0 safe */

  *combo = gtk_combo_box_entry_new_text();
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(*combo))->child);    /* the entry */

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), *combo, FALSE, FALSE, 0);

  gtk_entry_set_max_length(centry, entry_max_length);
  gtk_editable_set_editable(GTK_EDITABLE(centry), entry_is_editable);

  if (glist.items) {                                      /* fill combo list */
    g_list_foreach(glist.items, combo_box_append_text, (gpointer)(*combo));
  }

  if (entry_cb) {
    g_signal_connect(G_OBJECT(centry), "activate", G_CALLBACK(entry_cb), 
		     entry_cb_data);
  }

  return hbox;
}



/* Packs a drawing into a viewport.

Revision: 29.07.2005 E.Sack

Input : - pointer to vertical viewport
        - pointer to horizontal viewport
        - drawing size
        - pointer to mouse callback
        - pointer to exposure callback
Output: - pointer to the drawing
*/
GtkWidget *
horizontal_viewport(GtkWidget *v_viewport, GtkWidget **h_viewport, 
		    area_def drawing_size, GtkSignalFunc mouse_cb, 
		    GtkSignalFunc expose_cb)
{
  GtkWidget *drawing, *hbox;

  /* events set in the window not only in the drawing which can be smaller */
  *h_viewport = gtk_event_box_new();  /* only because of offering a window */

  if (mouse_cb) {                                            
    gtk_widget_set_events(*h_viewport, GDK_BUTTON_PRESS_MASK);/* set here! */
    g_signal_connect(*h_viewport, "button-press-event", 
	   	     G_CALLBACK(mouse_cb), NULL);
  }

  gtk_box_pack_start(GTK_BOX(v_viewport), *h_viewport, TRUE, TRUE, 0);

  /* put into the vertical viewport a horizontal viewport                  */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(*h_viewport), hbox);

  /* put into the horizontal viewport a drawing area                       */
  drawing = gtk_drawing_area_new();

  if (expose_cb) {                          /* expose only for the drawing */
    g_signal_connect(G_OBJECT(drawing), "expose_event", 
	   	     G_CALLBACK(expose_cb), NULL);
  }

  if (AREA_SIZE(drawing_size) > 0) {   /* size not always known in advance */
    gtk_widget_set_size_request(GTK_WIDGET(drawing), 
				drawing_size.width, drawing_size.height);
  }

  gtk_box_pack_start(GTK_BOX(hbox), drawing, FALSE, FALSE, 0);

  gtk_widget_show_all(*h_viewport);

  return drawing;
}



/* Creates a scrolled drawing.

Revision: 05.11.2004 E.Sack

Input : - pointer to scrolled window definition
        - drawing size
        - pointer to mouse callback
        - pointer to exposure callback
Output: - scrolled window definition updated
*/
GtkWidget *
scrolled_drawing(scroll_def *swin, area_def drawing_size, 
		 GtkSignalFunc mouse_cb, GtkSignalFunc expose_cb)
{
  GtkWidget *scrolled_window;

  /* create a scrolled window */
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);

  /* fetch scrollbars */
  swin->h_adj = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  swin->v_adj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window));

  /* put into the scrolled window a vertical box as its vertical viewport */
  swin->v_viewport = gtk_vbox_new(FALSE, 0);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window), swin->v_viewport);
  gtk_container_set_border_width(GTK_CONTAINER(swin->v_viewport), XPLAB_IMG_BORDER_SIZE);

  /* put into the vertical viewport a horizontal box as its horizontal viewport */
  swin->sdrawing = horizontal_viewport(swin->v_viewport, &swin->h_viewport,
		 		       drawing_size, mouse_cb, expose_cb);
  return scrolled_window;
}



/* Creates a common scrolling and zooming image window.

Revision: 29.07.2005 E.Sack

Input : - pointer to window title string
        - function for upper part of window
        - function for lower part of window
        - pointer to scroll definition
        - drawing area size
        - pointer to mouse callback
        - pointer to slider callback
        - pointer to exposure callback
        - pointer to delete callback
        - window height add-on
        - window size
Output: - pointer to window
Precondition: - all function pointers may be NULL
              - all callback pointers may be NULL
              - the drawing area size may be 0
              - the functions can add any widgets to the window container
Postcondition: - if window size area calculates to 0, default values are
                 used.
               - window size is always limited to a size depending of the
                 screen size.
Usage: - the exposure callback is used to create the image by 
         gdk_draw_rgb_image().
*/
GtkWidget *
xplab_image_window(const char *window_title, void vbox_top(GtkWidget *vbox), 
		   void vbox_bot(GtkWidget *vbox), scroll_def *swin, 
		   area_def drawing_size, GtkSignalFunc mouse_cb, 
		   GtkSignalFunc slider_cb, GtkSignalFunc expose_cb,
		   GtkSignalFunc delete_cb, int extra_height, area_def far)
{
  GtkWidget *top_win, *vbox;
  gint      window_width, window_height, fram;
  gfloat    scrn_covr = XPLAB_SCRN_COVR; /* avoid total screen coverage by image */

  if (AREA_SIZE(far) > 0) {
    window_width  = far.width;
    window_height = far.height;
  }
  else {
    fram = 2*XPLAB_IMG_BORDER_SIZE + XPLAB_SLIDER_WIDTH + 
           XPLAB_WIN_FRAME_WIDTH;

    window_width  = drawing_size.width + fram;
    window_height = drawing_size.height + fram + extra_height;
  }

  window_width = XPMIN((gint)((gfloat)gdk_screen_width()*scrn_covr), 
		       window_width);

  window_height = XPMIN((gint)((gfloat)gdk_screen_height()*scrn_covr), 
			window_height);

  top_win = gtk_window_new(GTK_WINDOW_TOPLEVEL);         /* the top window */
  gtk_window_set_title(GTK_WINDOW(top_win), window_title);
  gtk_window_set_default_size(GTK_WINDOW(top_win), window_width, window_height);

  /* put into top window a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(top_win), vbox);

  if (vbox_top) {
    vbox_top(vbox);                     /* put into vertical box a menubar */
  }

  /* put into vertical box a scrolled image */
  zoom_scroll_window(vbox, swin, drawing_size, mouse_cb, slider_cb, expose_cb);

  if (vbox_bot) {
    vbox_bot(vbox);                   /* put into vertical box a statusbar */
  }

  if (delete_cb) {                                 /* should be set always */
    g_signal_connect(G_OBJECT(top_win), "delete-event", G_CALLBACK(delete_cb),
		     NULL);
  }

  xplab_set_window_icon(top_win);                           /* window icon */

  return top_win;
}



/* Builds a scrolled window.

Revision: 29.07.2005 E.Sack

Input : - pointer to container
        - pointer to scroll definition
        - drawing area size
        - pointer to mouse callback
        - pointer to slider callback
        - pointer to exposure callback
Output: - container updated
*/
static void
zoom_scroll_window(GtkWidget *vbox, scroll_def *swin, area_def drawing_size,
		   GtkSignalFunc mouse_cb, GtkSignalFunc slider_cb, 
		   GtkSignalFunc expose_cb)
{
  char      changed[] = "changed";
  char      vchanged[] = "value_changed";
  GtkWidget *scrolled_window;

  /* put into box one scrolled drawing */
  scrolled_window = scrolled_drawing(swin, drawing_size, mouse_cb, expose_cb);
  gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);

  if (slider_cb) {              /* signals used for pixel magnitude change */
    g_signal_connect(G_OBJECT(swin->h_adj), changed, G_CALLBACK(slider_cb),
		     NULL);
    g_signal_connect(G_OBJECT(swin->h_adj), vchanged, G_CALLBACK(slider_cb), 
		     NULL);
    g_signal_connect(G_OBJECT(swin->v_adj), changed, G_CALLBACK(slider_cb), 
		     NULL);
    g_signal_connect(G_OBJECT(swin->v_adj), vchanged, G_CALLBACK(slider_cb),
		     NULL);
  }
}



/* Draws an image into a drawing.

Revision: 06.11.2004 E.Sack

Input : - pointer to un-zoomed image buffer
        - image buffer size
        - zoomed image buffer
        - zoom factor
        - pointer to horizontal adjustment
        - pointer to vertical adjustment
        - pointer to drawing
Output: - drawing updated
Precondition: - the adjustments shall be consistent to the drawing window.
Usage:  - dependent from the zoom factor, different image buffers are used.
          The splitting enables to limit a zoomed image just to its visual
          clipping buffer and therefore to save memory consumption.
*/
void
xplab_draw_image(col_def *rgbbuf, area_def rgbsize, pic_def magbuf, 
		 gfloat magnitude, GtkAdjustment *h_adj, 
		 GtkAdjustment *v_adj, GtkWidget *drawing)
{
  guchar *image_to_render = NULL;
  gint   width_to_render, height_to_render, x_start_to_render;
  gint   y_start_to_render;

  if (magnitude == XPLAB_MAGN_X1_VALUE || magbuf.col == NULL) {
    if (rgbbuf == NULL) {
      return;
    }

    image_to_render = (guchar *)rgbbuf;
    x_start_to_render = 0;
    y_start_to_render = 0;
    width_to_render = (gint)rgbsize.width;
    height_to_render = (gint)rgbsize.height;
  }
  else {      /* use a different image buffer containing only visible data */
    image_to_render = (guchar *)magbuf.col;
    x_start_to_render = XPMAX(0, (gint)h_adj->value-XPLAB_IMG_BORDER_SIZE);/* keep in clipping */
    y_start_to_render = XPMAX(0, (gint)v_adj->value-XPLAB_IMG_BORDER_SIZE);
    width_to_render = (gint)magbuf.ar.width;
    height_to_render = (gint)magbuf.ar.height;
  }

  gdk_draw_rgb_image(drawing->window,                 /* image destination */
		     drawing->style->fg_gc[GTK_STATE_NORMAL],
		     x_start_to_render, y_start_to_render, 
		     width_to_render, height_to_render,
		     GDK_RGB_DITHER_NONE,
		     image_to_render, width_to_render * 3);/* image source */
}



/* Scales down backing image.

Revision: 20.11.2004 E.Sack

Input : - pointer to source image buffer
        - source image buffer size
        - source image buffer's clipping start coordinates
        - scaling factor
        - destination image
Output: - backing image updated
Precondition: - destination image must be allocated defining clipping.
*/
void
scale_down(col_def *src_img, area_def src_size, gfloat src_start_x, 
	   gfloat src_start_y, gfloat magnitude, pic_def dst_img)
{
  pic_def asm_img;

  init_pic_def(&asm_img);

  src_start_x = XPMAX(0.0, src_start_x - (gfloat)XPLAB_IMG_BORDER_SIZE);
  src_start_y = XPMAX(0.0, src_start_y - (gfloat)XPLAB_IMG_BORDER_SIZE);

  src_start_x /= magnitude;                                     /* rescale */
  src_start_y /= magnitude;

  asm_img.ar = src_size;
  asm_img.col = src_img;                                /* assemble format */

  xplab_scale_down_image(asm_img, dst_img, src_start_x, src_start_y, 
			 magnitude);
}



/* Scales up backing image.

Revision: 11.12.2004 E.Sack

Input : - pointer to source image buffer
        - source image buffer size
        - source image buffer's clipping start coordinates
        - scaling factor
        - destination image
        - visible area values
        - allocation area values
Output: - backing image updated
Usage:  - the image buffer displayed is ment to be hold in a scrolled
          window area.
*/
static void
scale_up(col_def *src_img, area_def src_size, gfloat src_start_x, 
	 gfloat src_start_y, gfloat magnitude, pic_def dst_img, 
	 gfloat visible_x, gfloat visible_y, int allocation_x, 
	 int allocation_y)
{
  int    netto_x, netto_y;
  gfloat start_x, start_y;

  pic_def asm_img;

  init_pic_def(&asm_img);

  start_x = XPMAX(0.0, src_start_x-(gfloat)XPLAB_IMG_BORDER_SIZE);
  start_y = XPMAX(0.0, src_start_y-(gfloat)XPLAB_IMG_BORDER_SIZE);

  start_x /= magnitude;                               /* rescale to source */
  start_y /= magnitude;

  netto_x = (int)net_displayed(src_start_x, visible_x, allocation_x, 
			       XPLAB_IMG_BORDER_SIZE);

  netto_y = (int)net_displayed(src_start_y, visible_y, allocation_y, 
			       XPLAB_IMG_BORDER_SIZE);

  netto_x = XPMIN(netto_x, dst_img.ar.width);
  netto_y = XPMIN(netto_y, dst_img.ar.height);

  asm_img.ar = src_size;
  asm_img.col = src_img;                                /* assemble format */

  xplab_scale_up_image(asm_img, dst_img, start_x, start_y, magnitude, 
		       netto_x, netto_y);
}



/* Sets up the appropriate integer scaled up/down backing image.

Revision: 28.11.2004 E.Sack

Input : - pointer to source image buffer
        - source image buffer size
        - source image buffer's clipping start coordinates
        - scaling factor
        - destination image
        - visible area values
        - allocation area values
        - flag (0: don't redraw always, 1: redraw always)
Output: - backing image updated
*/
static void
magnitude_content(col_def *src_img, area_def src_size, gfloat src_start_x, 
		  gfloat src_start_y, gfloat magnitude, pic_def *dst_img, 
		  gfloat visible_x, gfloat visible_y, int allocation_x, 
		  int allocation_y, gint force)
{
  if (magnitude == XPLAB_MAGN_X1_VALUE) {
    return;
  }

  assert(dst_img->col);

  if (force ||                                    /* new content requested */
      (dst_img->org.x != (int)src_start_x) ||
      (dst_img->org.y != (int)src_start_y)) {
    dst_img->org.x = (int)src_start_x;
    dst_img->org.y = (int)src_start_y;

    if (magnitude > XPLAB_MAGN_X1_VALUE) {
      scale_up(src_img, src_size, src_start_x, src_start_y, magnitude, 
	       *dst_img, visible_x, visible_y, allocation_x, allocation_y);
    }
    else {
      scale_down(src_img, src_size, src_start_x, src_start_y, magnitude,
		 *dst_img);
    }
  }
}



/* Sets up the appropriate image buffer for display.

Revision: 29.07.2005 E.Sack

Input : - pointer to source image buffer
        - source image buffer size
        - source image buffer's clipping start coordinates
        - scaling factor
        - destination image
        - visible area values
        - allocation area values
        - flag (0: don't redraw always, 1: redraw always)
Output: - backing image updated
Usage:  - apart from the processed image RGB, which always has the original
          size and is never scaled, an extra image buffer MAG is used to
          hold the displayed image data when the scale differs from 1. This
          is made to decouple the scaling factor from the displayed image
          buffer size. Hence, any factor greater than 0 is allowed and the
          MAG size depends only from the current amount of displayed image
          data. The MAG is deallocated completely if the factor is set to 1,
          in which case the RGB buffer is used directly for display.
*/
void
magnitude_allocation(col_def *src_img, area_def src_size, gfloat src_start_x, 
		     gfloat src_start_y, gfloat magnitude, pic_def *dst_img,
		     gfloat visible_x, gfloat visible_y, int allocation_x, 
		     int allocation_y, gint force_redraw)
{
  int    h_tmp, w_tmp;
  gfloat magn_width, magn_height;

  assert(magnitude>0.0);

  if (magnitude == XPLAB_MAGN_X1_VALUE) {
    free_pic_def(dst_img);                                   /* not needed */
    return;
  }

  magn_width =                              /* only visible area allocated */
    XPMIN((gfloat)src_size.width*magnitude, visible_x);

  magn_height = 
    XPMIN((gfloat)src_size.height*magnitude, visible_y);

  if (magn_width < 1.0 || magn_height < 1.0) {
    return;                          /* cannot display less than one pixel */
  }

  w_tmp = (int)(magn_width + 0.5);
  h_tmp = (int)(magn_height + 0.5);

  if (dst_img->col == NULL ||                        /* new size requested */
      (dst_img->ar.width != w_tmp) ||
      (dst_img->ar.height != h_tmp)) {
    dst_img->ar.width = w_tmp;
    dst_img->ar.height = h_tmp;

    if (dst_img->col) {
      REALLOC(col_def, AREA_SIZE(dst_img->ar), dst_img->col);
    }
    else {
      CALLOC(col_def, AREA_SIZE(dst_img->ar), dst_img->col);
    }

    force_redraw = 1;                      /* new allocation to be updated */
  }

  magnitude_content(src_img, src_size, src_start_x, 
		    src_start_y, magnitude, dst_img, visible_x, visible_y, 
		    allocation_x, allocation_y, force_redraw);
}



/* Calculates visible interval of a virtual image contained in a window.

Revision: 27.11.2004 E.Sack

Input : - start of the image clipping
        - visible part of the drawing
        - allocated size of the drawing
        - border size of the image
Output: - net interval size needed for image data
Usage: - the image shall be displayed in a drawing area which again is
         contained in a scrolled window. The drawing area might be greater
         or lesser than the window size. This is true for the image, too.
         This function calculates only a 1-D interval and hence must be
         called twice to get 2-D image buffer sizes.
*/
static gfloat
net_displayed(gfloat start, gfloat visible, int allocation, int offset)
{
  gfloat net;
  gfloat offs = (gfloat)offset;                /* border size around image */
  gfloat allc = (gfloat)allocation;      /* allocated interval for display */

  net = (int)(visible                /* visible part of allocated interval */
	      - XPMAX(0.0, offs - start)                    /* left border */
	      - XPMAX(0.0, start + visible - allc - offs));/* right border */

  return net;                         /* current interval size for display */
}



/* Creates a menu entry.

Revision: 29.07.2005 E.Sack

Input : - pointer to menu
        - callback function
        - pointer to callback user data
Output: - menu updated
Usage: - a submenu list is created containing different zooming factors.
*/
void
xplab_menu_magn(GtkWidget *menu_widget, GtkSignalFunc mag_cb, 
		gpointer udata, gfloat mag_act)
{
  GSList    *group = NULL;
  GtkWidget *magnitude_entry, *submenu;

  magnitude_entry = gtk_menu_item_new_with_label(MENU_BUTTON_MAGN);

  assert(GTK_IS_MENU_SHELL(menu_widget) == TRUE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_widget), magnitude_entry);

  submenu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(magnitude_entry), submenu);

  magnitude_menu_item(XPLAB_MAGN_X1_VALUE, mag_act, &group, submenu, mag_cb, udata);
  magnitude_menu_item(XPLAB_MAGN_X2_VALUE, mag_act, &group, submenu, mag_cb, udata);
  magnitude_menu_item(XPLAB_MAGN_X4_VALUE, mag_act, &group, submenu, mag_cb, udata);
  magnitude_menu_item(XPLAB_MAGN_D2_VALUE, mag_act, &group, submenu, mag_cb, udata);
  magnitude_menu_item(XPLAB_MAGN_D4_VALUE, mag_act, &group, submenu, mag_cb, udata);

  gtk_widget_show_all(magnitude_entry);                 /* sometimes needed */
}



/* Adds an item to a menu.

Revision: 29.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
magnitude_menu_item(gfloat mag, gfloat mag_act, GSList **group, 
		    GtkWidget *magnitude_menu, GtkSignalFunc magnitude_cb,
		    gpointer udata)
{
  gchar     *buf = NULL;
  GtkWidget *magnitude_item;
  
  if (mag == XPLAB_MAGN_X1_VALUE) {
    add_string(XPLAB_MAGN_X1_NAME, &buf);
  }
  else if (mag == XPLAB_MAGN_X2_VALUE) {
    add_string(XPLAB_MAGN_X2_NAME, &buf);
  }
  else if (mag == XPLAB_MAGN_X4_VALUE) {
    add_string(XPLAB_MAGN_X4_NAME, &buf);
  }
  else if (mag == XPLAB_MAGN_D2_VALUE) {
    add_string(XPLAB_MAGN_D2_NAME, &buf);
  }
  else if (mag == XPLAB_MAGN_D4_VALUE) {
    add_string(XPLAB_MAGN_D4_NAME, &buf);
  }
  else {
    assert(0);                                  /* magnitude not supported */
  }

  magnitude_item = gtk_radio_menu_item_new_with_label(*group, buf);
  FREEMEM(buf);

  *group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(magnitude_item));

  if (mag == mag_act) {
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(magnitude_item), TRUE);
  }

  gtk_menu_shell_append(GTK_MENU_SHELL(magnitude_menu), magnitude_item);

  g_signal_connect(magnitude_item, "toggled", G_CALLBACK(magnitude_cb), 
		   udata);
}



/* Updates text store.

Revision: 06.08.2005 E.Sack

Input : - pointer to attribute
Output: - text store updated
*/
void
put_history_to_combo_store(entry_def *att)
{
  gint     err, nodes;
  gchar    *gbuf;
  GtkEntry *centry;

  assert(att);

  if (XPLAB_HISTORY_LENGTH <= 0) {
    return;                                                    /* no history */
  }

  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att->combo))->child);

  gbuf = xplab_entry_get_text(centry);

  err = check_name(gbuf);

  if (!err) {
    GtkTreeModel *tmodel;
    foreach_def  fdef;

    init_foreach_def(&fdef);

    tmodel = gtk_combo_box_get_model(GTK_COMBO_BOX(att->combo));

    fdef.iaux = XPLAB_COMBO_STORE_COLUMN_TXT;/* set combo store column index */
    fdef.txt = gbuf;           /* set compare string pointer, don't free it. */

    gtk_tree_model_foreach(tmodel,                /* investigate combo store */
			   (GtkTreeModelForeachFunc)tree_model_search_strings,
			   (gpointer)&fdef);

    if (fdef.baux == FALSE) {
      nodes = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(tmodel), NULL);

      if (nodes >= XPLAB_HISTORY_LENGTH) {
	gtk_combo_box_remove_text(GTK_COMBO_BOX(att->combo), 0);/*first entry*/
      }

      gtk_combo_box_append_text(GTK_COMBO_BOX(att->combo), gbuf);
    }
  }

  FREEMEM(gbuf);
}



/* Calculates the training image coordinates from the mouse pointer.

Revision: 14.01.2006 E.Sack

Input : - pointer to GDK event
        - pointer to coordinates
        - magnitude factor
Output: - coordinates updated
        - flag (0: coordinates not within image, 1: coordinates within image)
*/
int
image_coordinates(GdkEventButton *event, c_def *start, sector_def sec, 
		  float magnitude)
{
  if (magnitude <= 0) {
    return 0;                                    /* avoid division by zero */
  }

  start->x = (int)(event->x/magnitude);            /* original coordinates */
  start->y = (int)(event->y/magnitude);

  return WITHIN_SECTOR(start->x, start->y, sec);      /* inside image flag */
}
